import logging
import reprlib
from bisect import bisect_left, bisect_right
from itertools import chain
from operator import attrgetter

import numpy as np
import toolz
from pyquaternion import Quaternion

from ada.config import Settings

from .utils import Counter, points_in_cylinder, vector_length

__all__ = [
    "Nodes",
    "Beams",
    "Plates",
    "Connections",
    "Materials",
    "Sections",
]


class BaseCollections:
    """
    The Base class for all collections

    :param parent:
    """

    def __init__(self, parent):
        self._parent = parent


class Beams(BaseCollections):
    """
    A collections of Beam objects

    :param beams:
    :param unique_ids:
    :param parent:
    """

    def __init__(self, beams=None, unique_ids=True, parent=None):

        super().__init__(parent)
        beams = [] if beams is None else beams
        if unique_ids:
            beams = toolz.unique(beams, key=attrgetter("name"))
        self._beams = sorted(beams, key=attrgetter("name"))
        self._dmap = {n.name: n for n in self._beams}

    def __contains__(self, item):
        return item.guid in self._dmap.keys()

    def __len__(self):
        return len(self._beams)

    def __iter__(self):
        return iter(self._beams)

    def __getitem__(self, index):
        result = self._beams[index]
        return Beams(result) if isinstance(index, slice) else result

    def __eq__(self, other):
        if not isinstance(other, Beams):
            return NotImplemented
        return self._beams == other._beams

    def __ne__(self, other):
        if not isinstance(other, Beams):
            return NotImplemented
        return self._beams != other._beams

    def __add__(self, other):
        return Beams(chain(self._beams, other._beams))

    def __repr__(self):
        rpr = reprlib.Repr()
        rpr.maxlist = 8
        rpr.maxlevel = 1
        return f"Beams({rpr.repr(self._beams) if self._beams else ''})"

    def index(self, item):
        index = bisect_left(self._beams, item)
        if (index != len(self._beams)) and (self._beams[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def count(self, item):
        return int(item in self)

    def from_name(self, name):
        """
        Get beam from its name

        :param name: Beam name
        :return: Beam object
        """
        if name not in self._dmap.keys():
            raise ValueError(f'The beam "{name}" is not found')
        else:
            return self._dmap[name]

    def add(self, beam):
        """

        :param beam:
        :type beam: ada.Beam
        """
        if beam.name is None:
            raise Exception("Name is not allowed to be None.")

        if beam.name in self._dmap.keys():
            return self._dmap[beam.name]
        self._dmap[beam.name] = beam
        self._beams.append(beam)

    def get_beams_within_volume(self, vol_, margins=None):
        """
        :param vol_: List or tuple of tuples [(xmin, xmax), (ymin, ymax), (zmin, zmax)]
        :param margins: Add margins to the volume box (equal in all directions). Input is in meters. Can be negative.
        :return: List of beam ids
        """
        from bisect import bisect_left, bisect_right

        from .utils import roundoff

        if margins is not None:
            vol_new = []
            for p in vol_:
                vol_new.append((roundoff(p[0] - margins), roundoff(p[1] + margins)))
        else:
            vol_new = vol_
        vol = vol_new

        def sort_bms(bms):
            xkeys = [key[1] for key in bms]
            xmin = bisect_left(xkeys, vol[0][0])
            xmax = bisect_right(xkeys, vol[0][1])

            within_x_list = sorted(bms[xmin:xmax], key=lambda elem: elem[2])

            ykeys = [key[2] for key in within_x_list]
            ymin = bisect_left(ykeys, vol[1][0])
            ymax = bisect_right(ykeys, vol[1][1])

            within_y_list = sorted(within_x_list[ymin:ymax], key=lambda elem: elem[3])

            zkeys = [key[3] for key in within_y_list]
            zmin = bisect_left(zkeys, vol[2][0])
            zmax = bisect_right(zkeys, vol[2][1])

            within_vol_list = within_y_list[zmin:zmax]
            return [bm[0] for bm in within_vol_list]

        bm_list1 = [(bm.name, bm.n1.x, bm.n1.y, bm.n1.z) for bm in sorted(self._beams, key=lambda bm: bm.n1.x)]
        bm_list2 = [(bm.name, bm.n2.x, bm.n2.y, bm.n2.z) for bm in sorted(self._beams, key=lambda bm: bm.n2.x)]

        return list(set([self._dmap[bm_id] for bms_ in (bm_list1, bm_list2) for bm_id in sort_bms(bms_)]))

    @property
    def dmap(self):
        """

        :return: A dictionary of all beams {int(id1):node1, ..}
        :rtype: dict
        """
        return self._dmap


class Plates(BaseCollections):
    """
    Plate object collection

    :param plates:
    :param unique_ids:
    :param parent:
    """

    def __init__(self, plates=None, unique_ids=True, parent=None):

        plates = [] if plates is None else plates
        super().__init__(parent)

        if unique_ids:
            plates = toolz.unique(plates, key=attrgetter("name"))
        self._plates = sorted(plates, key=attrgetter("name"))
        self._dmap = {n.name: n for n in self._plates}

    def __contains__(self, item):
        return item.id in self._dmap.keys()

    def __len__(self):
        return len(self._plates)

    def __iter__(self):
        return iter(self._plates)

    def __getitem__(self, index):
        result = self._plates[index]
        return Materials(result) if isinstance(index, slice) else result

    def __eq__(self, other):
        if not isinstance(other, Plates):
            return NotImplemented
        return self._plates == other._plates

    def __ne__(self, other):
        if not isinstance(other, Plates):
            return NotImplemented
        return self._plates != other._plates

    def __add__(self, other):
        return Plates(chain(self._plates, other._plates))

    def __repr__(self):
        rpr = reprlib.Repr()
        rpr.maxlist = 8
        rpr.maxlevel = 1
        return f"Plates({rpr.repr(self._plates) if self._plates else ''})"

    def index(self, item):
        index = bisect_left(self._plates, item)
        if (index != len(self._plates)) and (self._plates[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def count(self, item):
        return int(item in self)

    def get_by_id(self, name):
        """
        Get plate by its name

        :param name: Name of plate
        :return: Plate object
        """
        if name not in self._dmap.keys():
            raise ValueError(f'The node id "{name}" is not found')
        else:
            return self._dmap[name]

    @property
    def dmap(self):
        """

        :return: A dictionary of all nodes {int(id1):node1, ..}
        :rtype: dict
        """
        return self._dmap

    def _add_node_to_global(self, pnt):
        """

        :param pnt:
        :type pnt: ada.Node
        """
        old_node = self._parent.nodes.add(pnt)
        if old_node is not None:
            pnt = old_node
        return pnt

    def add(self, plate):
        """

        :param plate:
        :type plate: ada.Plate
        """
        if plate.name is None:
            raise Exception("Name is not allowed to be None.")

        if plate.name in self._dmap.keys():
            return self._dmap[plate.name]
        mat = self._parent.materials.add(plate.material)
        if mat is not None:
            plate.material = mat

        self._plates.append(plate)


class Connections(BaseCollections):
    counter = Counter(1, "C")
    """
    Connections Collection.

    :param connections: List of connections
    :param parent: Parent object
    """

    def __init__(self, connections=None, parent=None):

        connections = [] if connections is None else connections
        super().__init__(parent)
        self._connections = connections
        self._dmap = {j.id: j for j in self._connections}

    def __contains__(self, item):
        return item.id in self._dmap.keys()

    def __len__(self):
        return len(self._connections)

    def __iter__(self):
        return iter(self._connections)

    def __getitem__(self, index):
        result = self._connections[index]
        return Materials(result) if isinstance(index, slice) else result

    def __eq__(self, other):
        if not isinstance(other, Beams):
            return NotImplemented
        return self._connections == other._beams

    def __ne__(self, other):
        if not isinstance(other, Beams):
            return NotImplemented
        return self._connections != other._beams

    def __add__(self, other):
        return Beams(chain(self._connections, other._beams))

    def __repr__(self):
        rpr = reprlib.Repr()
        rpr.maxlist = 8
        rpr.maxlevel = 1
        return f"Connections({rpr.repr(self._connections) if self._connections else ''})"

    def add(self, joint):
        """
        Add a joint

        :param joint:
        :type joint: ada.Connection
        """
        if joint.name is None:
            raise Exception("Name is not allowed to be None.")

        if joint.name in self._dmap.keys():
            return self._dmap[joint.name]

        self._connections.append(joint)

    def find(self, out_of_plane_tol=0.1, joint_func=None):
        """
        Find all connections between beams in all parts using a simple clash check.

        :param out_of_plane_tol:
        :param joint_func: Pass a function for mapping the generic Connection classes to a specific reinforced Joints
        """
        from ada import Beam, Connection
        from ada.core.utils import beam_cross_check

        ass = self._parent.get_assembly()
        bm_res = ass.beam_clash_check()

        point_tuples = []

        def are_beams_connected(beams):
            """

            :param beams: Tuple containing beam and list of beams found using clash check
            :return:
            """
            bm1 = beams[0]
            assert isinstance(bm1, Beam)
            cross_beams = dict()
            for bm2 in beams[1]:
                if bm1 == bm2:
                    continue
                assert isinstance(bm2, Beam)
                res = beam_cross_check(bm1, bm2, out_of_plane_tol)
                if res is None:
                    continue
                point, s, t = res
                self._eval_joint_ends(bm1, bm2, t, point)

                if point is not None:
                    tp = tuple(point)
                    if tp not in cross_beams.keys():
                        cross_beams[tp] = []
                    cross_beams[tp].append(bm2)

            for p, mem in cross_beams.items():
                if p in point_tuples:
                    continue
                point_tuples.append(p)
                if joint_func is not None:
                    joint = joint_func(next(self.counter), [bm1] + mem)
                    if joint is None:
                        continue
                else:
                    joint = Connection(next(self.counter), [bm1] + mem)

                bm1.connected_from.append(joint)
                for m in mem:
                    m.connected_to.append(joint)
                self.add(joint)

        list(map(are_beams_connected, bm_res))

    def _eval_joint_ends(self, bm1, bm2, t_, intersect_point):
        """
        Evaluate the  use AB_ to ensure that the node lands on the beam

        :param bm1:
        :param bm2:
        :param t_:
        :param intersect_point:
        :return:
        """

        from ada import Node

        def eval_node(ab, n1=True):
            n = Node(ab)
            n_old = self._parent.nodes.add(n)

            # TODO: Evaluate if the following is necessary!
            if n_old is not None:
                n = n_old
            if n.id not in [bm2.n1.id, bm2.n2.id]:
                if n1 is True:
                    bm2.n1 = n
                    bm2.n1.Free = False
                else:
                    bm2.n2 = n
                    bm2.n2.Free = False
            else:
                logging.debug("Midnode on n1")

        if t_ <= 0:
            eval_node(intersect_point, n1=True)
        elif t_ >= 1:
            eval_node(intersect_point, n1=False)
        else:
            logging.error('bm1 "{}", bm2 "{}", t: "{}"'.format(bm1.name, bm2.name, t_))


class Materials(BaseCollections):
    """
    Collection of materials

    :param materials:
    :param unique_ids:
    :param parent:
    """

    def __init__(self, materials=None, unique_ids=True, parent=None, units="m"):
        super().__init__(parent)
        self._materials = sorted(materials, key=attrgetter("name")) if materials is not None else []
        self._unique_ids = unique_ids
        self._dmap = {n.name: n for n in self._materials}
        self._idmap = {n.id: n for n in self._materials}
        self._units = units

    def __contains__(self, item):
        return item.name in self._dmap.keys()

    def __len__(self):
        return len(self._materials)

    def __iter__(self):
        return iter(self._materials)

    def __getitem__(self, index):
        # if index not in self._dmap.keys():
        #     raise ValueError(f'Material name "{index}" not found')
        result = self._materials[index]
        return Materials(result) if isinstance(index, slice) else result

    def __eq__(self, other):
        if not isinstance(other, Materials):
            return NotImplemented
        return self._materials == other._materials

    def __ne__(self, other):
        if not isinstance(other, Materials):
            return NotImplemented
        return self._materials != other._materials

    def __add__(self, other):
        return Materials(chain(self._materials, other._materials))

    def __repr__(self):
        rpr = reprlib.Repr()
        rpr.maxlist = 8
        rpr.maxlevel = 1
        return f"Materials({rpr.repr(self._materials) if self._materials else ''})"

    def index(self, item):
        """

        :param item:
        :type item: ada.Material
        :return:
        """
        return self._materials.index(item)

    def count(self, item):
        return int(item in self)

    def get_by_name(self, name):
        """

        :param name:
        :return:
        """
        if name not in self._dmap.keys():
            raise ValueError(f'The material name "{name}" is not found')
        else:
            return self._dmap[name]

    def get_by_id(self, mat_id):
        """

        :param mat_id:
        :return:
        """
        if mat_id not in self._idmap.keys():
            raise ValueError(f'The material id "{mat_id}" is not found')
        else:
            return self._idmap[mat_id]

    @property
    def dmap(self):
        """

        :return: A dictionary of all nodes {int(id1):node1, ..}
        :rtype: dict
        """
        return self._dmap

    @property
    def parent(self):
        """

        :return:
        :rtype: ada.Part
        """
        return self._parent

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if value != self._units:
            for m in self._materials:
                m.units = value
            self._units = value

    def add(self, material):
        """

        :param material:
        :type material: ada.Material
        """
        if material in self:
            return self._dmap[material.name]

        if material.id is None or material.id in self._idmap.keys():
            material.id = len(self._materials) + 1
        self._idmap[material.id] = material
        self._dmap[material.name] = material
        self._materials.append(material)


class Sections:
    """

    :param sections:
    :param unique_ids:
    :param parent:
    """

    def __init__(self, sections=None, unique_ids=True, parent=None):

        sections = [] if sections is None else sections
        self._parent = parent

        if unique_ids:
            sections = list(toolz.unique(sections, key=attrgetter("name")))
        self._sections = sorted(sections, key=attrgetter("name"))
        self._nmap = {n.name: n for n in self._sections}
        self._idmap = {n.id: n for n in self._sections}
        if len(self._nmap.keys()) != len(self._idmap.keys()):
            raise ValueError("Non-unique ids or name are observed..")

    def __contains__(self, item):
        return item.id in self._nmap.keys()

    def __len__(self):
        return len(self._sections)

    def __iter__(self):
        return iter(self._sections)

    def __getitem__(self, index):
        result = self._sections[index]
        return Sections(result) if isinstance(index, slice) else result

    def __add__(self, other):
        return Sections(chain(self._sections, other._sections))

    def __repr__(self):
        rpr = reprlib.Repr()
        rpr.maxlist = 8
        rpr.maxlevel = 1
        return f"Sections({rpr.repr(self._sections) if self._sections else ''})"

    def _change_units(self, scale_factor):
        for s in self._sections:
            s._change_units(scale_factor)

    def index(self, item):
        index = bisect_left(self._sections, item)
        if (index != len(self._sections)) and (self._sections[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def count(self, item):
        return int(item in self)

    def get_by_name(self, name):
        """

        :param name: Name of id
        :rtype: ada.Section
        """
        if name not in self._nmap.keys():
            raise ValueError(f'The section id "{name}" is not found')
        else:
            return self._nmap[name]

    def get_by_id(self, sec_id):
        """

        :param sec_id: Section ID
        :rtype: ada.Section
        """
        if sec_id not in self._idmap.keys():
            raise ValueError(f'The node id "{sec_id}" is not found')
        else:
            return self._idmap[sec_id]

    @property
    def nmap(self):
        """

        :return: A dictionary of all nodes {int(id1):node1, ..}
        :rtype: dict
        """
        return self._nmap

    @property
    def idmap(self):
        """

        :return: A dictionary of all nodes {name:node1, ..}
        :rtype: dict
        """
        return self._idmap

    def add(self, section):
        """

        :param section:
        :type section: ada.Section
        """
        if section.name is None:
            raise Exception("Name is not allowed to be None.")

        # Note: Evaluate if parent should be "Sections" not Part object?
        if section.parent is None:
            section.parent = self._parent

        if section.name in self._nmap.keys():
            return self._nmap[section.name]

        if section.id is None or section.id in self._idmap.keys():
            new_sec_id = len(self._sections) + 1
            section.edit(sec_id=new_sec_id)

        self._sections.append(section)
        self._idmap[section.id] = section
        self._nmap[section.name] = section


class Nodes:
    """

    :param nodes:
    :param unique_ids:
    :param parent:
    :param from_np_array:
    """

    def __init__(self, nodes=None, unique_ids=True, parent=None, from_np_array=None):
        self._parent = parent
        if from_np_array is not None:
            self._array = from_np_array
            nodes = self._np_array_to_nlist(from_np_array)
        else:
            nodes = [] if nodes is None else nodes

        if unique_ids is True:
            nodes = toolz.unique(nodes, key=attrgetter("id"))

        self._nodes = sorted(nodes, key=attrgetter("x", "y", "z"))
        self._idmap = {n.id: n for n in sorted(self._nodes, key=attrgetter("id"))}
        self._maxid = max(self._idmap.keys()) if len(self._nodes) > 0 else 0
        self._bbox = self._get_bbox() if len(self._nodes) > 0 else None

    def renumber(self):
        """
        Ensures that the node numberings starts at 1 and has no holes in its numbering.

        """
        for i, n in enumerate(sorted(self._nodes, key=attrgetter("id")), start=1):
            if i != n.id:
                n.id = i

        self._idmap = {n.id: n for n in sorted(self._nodes, key=attrgetter("id"))}
        self._maxid = max(self._idmap.keys()) if len(self._nodes) > 0 else 0
        self._bbox = self._get_bbox() if len(self._nodes) > 0 else None

    def _np_array_to_nlist(self, np_array):
        from ada import Node

        return [Node(row[1:], int(row[0]), parent=self._parent) for row in np_array]

    def nlist_to_np_array(self, nlist):
        return np.array([(n.id, *n) for n in nlist])

    def __contains__(self, item):
        return item.id in self._idmap.keys()

    def __len__(self):
        return len(self._nodes)

    def __iter__(self):
        return iter(self._nodes)

    def __getitem__(self, index):
        result = self._nodes[index]
        return Nodes(result) if isinstance(index, slice) else result

    def __eq__(self, other):
        if not isinstance(other, Nodes):
            return NotImplemented
        return self._nodes == other._nodes

    def __ne__(self, other):
        if not isinstance(other, Nodes):
            return NotImplemented
        return self._nodes != other._nodes

    def __add__(self, other):
        return Nodes(chain(self._nodes, other._nodes))

    def __repr__(self):
        return f"Nodes({len(self._nodes)}, max_id: {self.max_nid}, min_id: {self.min_nid})"

    def index(self, item):
        index = bisect_left(self._nodes, item)
        if (index != len(self._nodes)) and (self._nodes[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def count(self, item):
        return int(item in self)

    def move(self, move=None, rotate=None):
        """
        A method for translating and/or rotating your model.

        :param move: Translate Nodes by parsing a Numpy array containing dX,dY,dZ vector np.array([dx,dy,dz])
        :param rotate: Rotate Nodes around a specified axis before translation.
                       Input is a tuple([p1X, p1Y, p1Z], [p2X, p2Y, p2Z], degrees)
        """

        def moving(no):
            no.p = no.p + move

        def map_rotations(no, p):
            no.p = p

        if rotate is not None:
            p1 = np.array(rotate[0])
            p2 = np.array(rotate[1])
            deg = rotate[2]
            my_quaternion = Quaternion(axis=p2 - p1, degrees=deg)
            rot_mat = my_quaternion.rotation_matrix
            vectors = np.array([n.p - p1 for n in self._nodes])
            res = np.matmul(vectors, np.transpose(rot_mat))
            [map_rotations(n, p + p1) for n, p in zip(self._nodes, res)]

        if move is not None:
            move = np.array(move)
            list(map(moving, self._nodes))

        self._nodes = sorted(self._nodes, key=attrgetter("x", "y", "z"))

    def from_id(self, nid):
        """

        :param nid:
        :return:
        """
        if nid not in self._idmap.keys():
            raise ValueError(f'The node id "{nid}" is not found')
        else:
            return self._idmap[nid]

    def _get_bbox(self):
        if len(self._nodes) == 0:
            raise ValueError("No Nodes are found")
        nodes_yids = sorted(self._nodes, key=attrgetter("y"))
        nodes_zids = sorted(self._nodes, key=attrgetter("z"))
        xmin, xmax = self._nodes[0], self._nodes[-1]
        ymin, ymax = nodes_yids[0], nodes_yids[-1]
        zmin, zmax = nodes_zids[0], nodes_zids[-1]
        return (xmin, xmax), (ymin, ymax), (zmin, zmax)

    @property
    def dmap(self):
        """

        :return: A dictionary of all nodes {int(id1):node1, ..}
        :rtype: dict
        """
        return self._idmap

    @property
    def bbox(self):
        """

        :return:
        """
        if self._bbox is None:
            self._bbox = self._get_bbox()
        return self._bbox

    @property
    def vol_cog(self):
        return tuple([(self.bbox[i][0][i] + self.bbox[i][1][i]) / 2 for i in range(3)])

    @property
    def max_nid(self):
        return max(self.dmap.keys()) if len(self.dmap.keys()) > 0 else 0

    @property
    def min_nid(self):
        return min(self.dmap.keys()) if len(self.dmap.keys()) > 0 else 0

    def get_by_volume(self, p=None, vol_box=None, vol_cyl=None, tol=1e-4):
        """

        :param p: Point
        :param vol_box: Additional point to find nodes inside a rectangular box
        :param vol_cyl: (radius, height, cylinder thickness). Note! Radius is measured to outside of cylinder wall
        :param tol: Point tolerance
        :return:
        """
        from ada import Node

        p = np.array(p) if type(p) is (list, tuple) else p
        if p is not None and vol_cyl is None and vol_box is None:
            vol = [(coord - tol, coord + tol) for coord in p]
        elif vol_box is not None:
            vol = list(zip(p, vol_box))
        elif vol_cyl is not None and p is not None:
            r, h, t = vol_cyl
            vol = [
                (p[0] - r - tol, p[0] + r + tol),
                (p[1] - r - tol, p[1] + r + tol),
                (p[2] - tol, p[2] + tol + h),
            ]
        else:
            raise Exception("No valid search input provided. None is returned")

        vol_min, vol_max = zip(*vol)
        xmin = bisect_left(self._nodes, Node(vol_min))
        xmax = bisect_right(self._nodes, Node(vol_max))

        xlist = sorted(self._nodes[xmin:xmax], key=attrgetter("y"))
        ysorted = [n.y for n in xlist]
        ymin = bisect_left(ysorted, vol_min[1])
        ymax = bisect_right(ysorted, vol_max[1])

        ylist = sorted(xlist[ymin:ymax], key=attrgetter("z"))
        zsorted = [n.z for n in ylist]
        zmin = bisect_left(zsorted, vol_min[2])
        zmax = bisect_right(zsorted, vol_max[2])

        simplesearch = ylist[zmin:zmax]

        if vol_cyl is not None:
            r, h, t = vol_cyl
            pt1_ = p + np.array([0, 0, -h])
            pt2_ = p + np.array([0, 0, +h])

            def eval_p_in_cyl(no):
                if t == r:
                    if points_in_cylinder(pt1_, pt2_, r, no.p) is True:
                        return no
                else:
                    eval1 = points_in_cylinder(pt1_, pt2_, r + t, no.p)
                    eval2 = points_in_cylinder(pt1_, pt2_, r - t, no.p)
                    if eval1 is True and eval2 is False:
                        return no
                return None

            return list(filter(None, [eval_p_in_cyl(q) for q in simplesearch]))
        else:
            return list(simplesearch)

    def add(self, node, point_tol=Settings.point_tol, allow_coincident=False):
        """
        Insert node into sorted list.

        :param node: Node object
        :param point_tol: Point Tolerance
        :param allow_coincident: Disable check for coincident nodes
        :type node: ada.Node
        """

        def insert_node(n, i):
            new_id = self._maxid + 1 if len(self._nodes) > 0 else 1
            if n.id in self._idmap.keys() or n.id is None:
                n.id = new_id

            self._nodes.insert(i, n)
            self._idmap[n.id] = n
            self._bbox = None
            self._maxid = n.id if n.id > self._maxid else self._maxid
            # self._bbox = self._get_bbox()

        index = bisect_left(self._nodes, node)

        if (index != len(self._nodes)) and (self._nodes[index] == node) and allow_coincident is False:
            old_node = self._nodes[index]
            vlen = vector_length(old_node.p - node.p)
            if vlen < point_tol:
                logging.debug(f'Replaced new node with node id "{self._nodes[index].id}" found within point tolerances')
                return self._nodes[index]
            else:
                insert_node(node, index)
        else:
            insert_node(node, index)

        return None



# coding=utf-8
import os
import pathlib
import shutil
import subprocess


class Blender(object):
    """
    A class wrapping around Blender


    Assuming you have added blender exe dir to system path, this will automatically pick it up. Otherwise, pass it as
    a variable

    :param work_dir:
    :param project_name:
    :param exe_path:
    """

    __exe_path = shutil.which("blender")

    def __init__(self, work_dir, project_name, exe_path=None):

        if self.__exe_path is None and exe_path is None:
            raise Exception(
                'Blender installation is not found. Either add it to system path env or pass the "exe_path"'
            )

        self.work_dir = work_dir
        self._project_name = project_name
        self.exe_path = self.__exe_path if exe_path is None else pathlib.Path(exe_path)

        # from ada.core.utils import getfileprop
        # self._bl_ver = getfileprop(self.exe_path)['FileVersion']
        # if os.path.isdir(self.project_path) is False:
        #     os.makedirs(self.project_path)
        # if os.path.isdir(self.assets_path) is False:
        #     os.makedirs(self.assets_path)
        # project_file = self.project_path + '\\Assembly.blend'
        # self._py_script = 'import bpy\n'
        # # self._py_script += 'bpy.context.user_preferences.view.show_splash = False\n'
        # if os.path.isfile(project_file):
        #     if self._check_if_blender_file_version_match(project_file, self._bl_ver):
        #         print('Existing "Assembly.Blend" file found. Opening file')
        #         self._py_script += 'bpy.ops.wm.open_mainfile(filepath=r"{}")'.format(project_file)
        # else:
        #     self._py_script += 'bpy.ops.wm.save_as_mainfile(filepath=r"{}")'.format(project_file)
        #     print('No blend file found. Creating a new file "Assembly.blend"')

    def run(self, script_path=None, run_silent=True):
        """
        Start Blender in the currently defined project_path


        """

        os.makedirs(self.work_dir, exist_ok=True)

        args = [str(self.exe_path)]
        if run_silent:
            args += ["-b"]

        if script_path is not None:
            script_path = pathlib.Path(script_path).resolve().absolute()

            if script_path.is_file() is False:
                raise FileNotFoundError("Script not found")
            args += ["--python", str(script_path)]

        cmdstring = tuple(args)

        subprocess.run(cmdstring, stdin=subprocess.PIPE, cwd=self.work_dir, shell=True)

    @property
    def scripts_dir(self):
        return self.work_dir / "scripts"

    @property
    def assets_dir(self):
        return self.work_dir / "assets"

    @property
    def work_dir(self):
        return self._work_dir

    @work_dir.setter
    def work_dir(self, value):
        self._work_dir = pathlib.Path(value).resolve().absolute()

    @property
    def exe_path(self):
        return self._exe_path

    @exe_path.setter
    def exe_path(self, value):
        self._exe_path = pathlib.Path(value).resolve().absolute()

    @property
    def project_name(self):
        return self._project_name



import re


class AbaFF:
    """
    Abaqus Fortran Flags. A class designed to aid in building regex searched
    for Abaqus flags.

    """

    def __init__(self, flag, args, subflags=None, nameprop=None):
        """

        :param flag: Main flag. *<FLAG>
        :param args: arguments. Tuple of arguments
        :param subflags:
        :param nameprop:


        type of arguments:

        >: up to

        """
        self._flag = flag
        self._subflags = subflags if subflags is not None else []
        self._nameprop = nameprop
        self._args = args

    def _create_regex(self, flag, nameprop, args):
        regstr = ""
        if nameprop is not None:
            regstr += rf"\*\*\s*{nameprop[0]}:\s*(?P<{nameprop[1]}>.*?)\n"
        regstr += rf"\*{flag}"
        for i, arg in enumerate(args):
            for j, fl in enumerate(arg):
                subfl = fl.replace("=", "").replace("|", "").replace(">", "")
                exact = True if "==" in fl else False
                if exact:
                    equality = False
                else:
                    equality = True if "=" in fl else False
                optional = True if "|" in fl else False
                uptostar = True if ">" in fl else False

                last_char = regstr[-10:]
                regstr += "(?:"

                if r"(?:\n|)\s*" != last_char:
                    regstr += ","

                clean_name = subfl.replace(" ", "_").replace(r"\*", "")

                if equality:
                    regstr += rf"\s*{subfl}=(?P<{clean_name}>"
                else:
                    regstr += rf"\s*(?P<{clean_name}>"

                if uptostar:
                    regstr += r"(?:(?!\*).)*"
                else:
                    if exact is True:
                        regstr += subfl
                    else:
                        regstr += ".*?"
                regstr += ")"

                if j + 1 == len(arg):
                    regstr += "$"

                if optional:
                    regstr += "|"
                    if j + 1 == len(arg):
                        regstr += "$"

                regstr += ")"
            regstr += r"(?:\n|)\s*"
        return regstr

    @property
    def regstr(self):
        regstr = self._create_regex(self._flag, self._nameprop, self._args)
        for v in self._subflags:
            clean_name = v[0].replace(" ", "_").replace(r"\*", "")
            regstr += f"(:?(?P<{clean_name}>)"
            regstr += self._create_regex(v[0], None, v[1])
            regstr += ")"
        return regstr

    @property
    def regex(self):
        return re.compile(self.regstr, re.IGNORECASE | re.MULTILINE | re.DOTALL)


class AbaCards:
    # Contact
    contact_pairs = AbaFF(
        "Contact Pair",
        [
            (
                "interaction=",
                "small sliding==|",
                "type=|",
                "adjust=|",
                "mechanical constraint=|",
                "geometric correction=|",
                "cpset=|",
            ),
            ("surf1", "surf2"),
        ],
        nameprop=("Interaction", "name"),
    )

    contact_general = AbaFF(
        "Contact",
        args=[()],
        subflags=[
            ("Contact Inclusions", [(), ("surf1", "surf2")]),
            ("Contact Property Assignment", [(), ("vara", "varb", "interaction")]),
            ("Contact Formulation", [("type=",), ("csurf1", "csurf2", "csurf_type")]),
            ("Contact Initialization Assignment", [(), ("ssurf1", "ssurf2", "cinit")]),
            ("Surface Property Assignment", [("property=",), ("bulk>",)]),
        ],
        # nameprop=("Interaction", "name"),
    )

    # Connectors
    connector_behaviour = AbaFF(
        "Connector Behavior",
        [("name=",)],
        [("Connector Elasticity", [("nonlinear|", "component=|"), ("bulk>",)])],
    )
    connector_section = AbaFF("Connector Section", [("elset=", "behavior="), ("contype",), ("csys",)])

    # Constraints
    sh2so_re = AbaFF("Shell to Solid Coupling", [("constraint name=",), ("surf1", "surf2")])
    rigid_bodies = AbaFF("Rigid Body", [("ref node=", "elset=")])
    coupling = AbaFF(
        "Coupling",
        [("constraint name=", "ref node=", "surface=", "orientation=|")],
        [("Kinematic", [(), ("bulk>",)])],
    )
    tie = AbaFF("Tie", [("name=", "adjust="), ("surf1", "surf2")])
    # Other
    surface_smoothing = AbaFF("Surface Smoothing", [("name=",), ("bulk>",)])
    surface = AbaFF("Surface", [("type=", "name=", "internal|"), ("bulk>",)])
    orientation = AbaFF(
        "Orientation",
        [
            ("name=", "definition=|", "local directions=|", "system=|"),
            ("ax", "ay", "az", "bx", "by", "bz", "|cx", "|cy", "|cz"),
            ("v1", "v2"),
        ],
    )



import os
import pathlib
import re
import shutil
import subprocess
import sys
import time
from contextlib import contextmanager

from ada.config import Settings as _Settings

try:
    # Python 3.6+
    from os import PathLike
except ImportError:
    from pathlib import PurePath as PathLike


class LocalExecute:
    """

    Backend Component for executing local analysis
    """

    def __init__(
        self,
        inp_path,
        cpus=2,
        gpus=None,
        run_ext=False,
        metadata=None,
        auto_execute=True,
        excute_locally=True,
    ):
        self._inp_path = pathlib.Path(inp_path)
        self._cpus = cpus
        self._gpus = gpus
        self._run_ext = run_ext
        self._metadata = metadata
        self._auto_execute = auto_execute
        self._local_execute = excute_locally

    def _run_local(self, run_command, stop_command=None, exit_on_complete=True):

        if sys.platform == "linux" or sys.platform == "linux2":
            run_linux(self, run_command)
        elif sys.platform == "darwin":
            run_macOS(self, run_command)
        else:  # sys.platform == "win32":
            run_windows(self, run_command, stop_command, exit_on_complete)

    def run(self):
        raise NotImplementedError("The run function is not implemented")

    @property
    def analysis_dir(self):
        return self.inp_path.parent

    @property
    def execute_dir(self):
        if _Settings.execute_dir is None:
            return self.analysis_dir
        else:
            return _Settings.execute_dir / self.analysis_name

    @property
    def analysis_name(self):
        return self.inp_path.stem

    @property
    def inp_path(self):
        return pathlib.Path(self._inp_path)


def is_buffer(obj, mode):
    return ("r" in mode and hasattr(obj, "read")) or ("w" in mode and hasattr(obj, "write"))


@contextmanager
def open_file(path_or_buf, mode="r"):
    if is_buffer(path_or_buf, mode):
        yield path_or_buf
    elif sys.version_info < (3, 6) and isinstance(path_or_buf, PathLike):
        # TODO remove when python 3.5 is EoL (i.e. 2020-09-13)
        # https://devguide.python.org/#status-of-python-branches
        # https://www.python.org/dev/peps/pep-0478/
        with open(str(path_or_buf), mode) as f:
            yield f
    else:
        with open(path_or_buf, mode) as f:
            yield f


def get_fem_model_from_assembly(assembly):
    """
    Scans the assembly tree for part (singular) containing FEM elements. Multiple parts with elements are not allowed

    :param assembly:
    :return: A single or multiple parts
    :rtype: ada.Part
    """
    parts = list(filter(lambda p: len(p.fem.elements) != 0, assembly.get_all_parts_in_assembly(True)))

    if len(parts) > 1:
        raise ValueError(
            "This method does not yet support multipart FEM. Please make sure your assembly only contain 1 FEM"
        )
    elif len(parts) == 0:
        raise ValueError("At least 1 part must have a FEM mesh ")

    return parts[0]


def get_exe_path(exe_name):
    """

    :param exe_name:
    :return:
    """
    import shutil

    from ada.config import Settings

    if Settings.fem_exe_paths[exe_name]:
        exe_path = Settings.fem_exe_paths[exe_name]
    elif os.getenv(f"ADA_{exe_name}_exe"):
        exe_path = os.getenv(f"ADA_{exe_name}_exe")
    elif shutil.which(f"{exe_name}.exe"):
        exe_path = shutil.which(f"{exe_name}.exe")
    elif shutil.which(f"{exe_name}.bat"):
        exe_path = shutil.which(f"{exe_name}.bat")
    else:
        raise FileNotFoundError(f'No Path to Executable "{exe_name}.exe" or "{exe_name}.bat" is found')

    exe_path = pathlib.Path(exe_path)

    if exe_path.exists() is False:
        raise FileNotFoundError(exe_path)

    return exe_path


def str_to_int(s):
    try:
        return int(float(s))
    except ValueError:
        raise ValueError("stop a minute")


def str_to_float(s):
    from ada.core.utils import roundoff

    return roundoff(s)


def get_ff_regex(flag, *args):
    """
    Compile a regex search string for Fortran formatted string input.

    :param flag: Name of keyword flag (ie. the first word on a line of input parameters)
    :param args: Group name for each parameter. Include the character | to signify the parameters is optional.


    :return: Returns a compiled regex search string.. re.compile(..)
    """
    pattern_str = r"^(?P<flag>.*?)" if flag is True else rf"^{flag}"
    counter = 0
    re_in = re.IGNORECASE | re.MULTILINE | re.DOTALL

    def add_key(k):
        if "|" in k:
            return rf'(?: \s*(?P<{k.replace("|", "")}>.*?)|)'
        return rf" \s*(?P<{k}>.*?)"

    for i, key in enumerate(args):
        if type(key) is str:
            pattern_str += add_key(key)
            counter += 1
            if counter == 4 and i < len(args) - 1:
                counter = 0
                pattern_str += r"(?:\n|)\s*"
        elif type(key) is list:
            for subkey in key:
                pattern_str += add_key(subkey)
            pattern_str += r"(?:\n|)\s*"
        else:
            raise ValueError(f'Unrecognized input type "{type(key)}"')

    if not args:
        pattern_str += add_key("bulk")

    return re.compile(pattern_str + r"(?:(?=^[A-Z]|\Z))", re_in)


def _overwrite_dir(analysis_dir):

    print("Removing old files before copying new")
    try:

        shutil.rmtree(analysis_dir)
    except WindowsError as e:
        print("Failed to delete due to '{}'".format(e))  # Or just pass

    time.sleep(0.5)
    os.makedirs(analysis_dir, exist_ok=True)


def _lock_check(analysis_dir):
    lck_file = (analysis_dir / analysis_dir.stem).with_suffix(".lck")
    if lck_file.is_file():
        raise IOError(
            f'Found lock-file:\n\n"{lck_file}"\n\nThis indicates that an analysis is running.'
            "Please stop analysis and try again"
        )
    if (analysis_dir / "ada.lck").is_file():
        raise IOError(
            f'Found ADA lock-file:\n\n"{analysis_dir}\\ada.lck"\n\nThis indicates that the analysis folder is'
            f" locked Please removed ADA lock file if this is not the case and try again"
        )


def _folder_prep(scratch_dir, analysis_name, overwrite):

    if scratch_dir is None:
        scratch_dir = pathlib.Path(_Settings.scratch_dir)
    else:
        scratch_dir = pathlib.Path(scratch_dir)

    analysis_dir = scratch_dir / analysis_name
    if analysis_dir.is_dir():
        _lock_check(analysis_dir)
        if overwrite is True:
            _overwrite_dir(analysis_dir)
        else:
            raise IOError("The analysis folder exists. Please remove folder and try again")

    os.makedirs(analysis_dir, exist_ok=True)
    return analysis_dir


def run_windows(exe, run_command, stop_command=None, exit_on_complete=True):
    """

    :param exe:
    :param run_command:
    :param stop_command:
    :param exit_on_complete:
    :return:
    """
    bat_start_str = f"""echo OFF
for %%* in (.) do set CurrDirName=%%~nx*
title %CurrDirName%
cd /d {exe.analysis_dir}
echo ON\ncall {run_command}"""

    if exit_on_complete is False:
        bat_start_str += "\npause"

    start_bat = "run.bat"
    stop_bat = "stop.bat"

    os.makedirs(exe.execute_dir, exist_ok=True)

    with open(exe.execute_dir / start_bat, "w") as d:
        d.write(bat_start_str + "\nEXIT")

    if stop_command is not None:
        with open(exe.execute_dir / stop_bat, "w") as d:
            d.write(f"cd /d {exe.analysis_dir}\n{stop_command}")

    if _Settings.execute_dir is not None:
        shutil.copy(exe.execute_dir / start_bat, _Settings.execute_dir / start_bat)
        shutil.copy(exe.execute_dir / stop_bat, _Settings.execute_dir / stop_bat)

    fem_tool = type(exe).__name__

    print(80 * "-")
    print(f'starting {fem_tool} simulation "{exe.analysis_name}"')
    if exe._auto_execute is True:
        if exe._run_ext is True:
            subprocess.run(
                "start " + start_bat,
                cwd=exe.execute_dir,
                shell=True,
                env=os.environ,
            )
            print(f"Note! This starts {fem_tool} in an external window on a separate thread.")
        else:
            subprocess.run(
                "start /wait " + start_bat,
                cwd=exe.execute_dir,
                shell=True,
                env=os.environ,
            )
            print(f'Finished {fem_tool} simulation "{exe.analysis_name}"')
    print(80 * "-")


def run_linux(exe, run_command):
    """

    :param exe:
    :param run_command:
    :return:
    """
    fem_tool = type(exe).__name__

    print(80 * "-")
    print(f'starting {fem_tool} simulation "{exe.analysis_name}" (on Linux)')
    if exe._auto_execute is True:
        if exe._run_ext is True:
            subprocess.run(
                run_command,
                cwd=exe.execute_dir,
                shell=True,
                env=os.environ,
            )
            print(f"Note! This starts {fem_tool} in an external window on a separate thread.")
        else:
            subprocess.run(
                run_command,
                cwd=exe.execute_dir,
                shell=True,
                env=os.environ,
            )
            print(f'Finished {fem_tool} simulation "{exe.analysis_name}"')
    print(80 * "-")


def run_macOS(exe, run_command):
    raise NotImplementedError()


def interpret_fem(fem_ref):
    fem_type = None
    if ".fem" in str(fem_ref).lower():
        fem_type = "sesam"
    elif ".inp" in str(fem_ref).lower():
        fem_type = "abaqus"
    return fem_type



import logging

from ada.config import Settings
from ada.fem.io.utils import get_exe_path

from ..utils import LocalExecute


def write_to_log(res_str, fname):
    with open(f'timeit_{fname.split(".")[0]}.log', "a") as d:
        d.write("\n" + res_str)


def run_code_aster(
    inp_path,
    cpus=2,
    gpus=None,
    run_ext=False,
    metadata=None,
    execute=True,
    return_bat_str=False,
    exit_on_complete=True,
):
    """

    TODO: Setup running for the code_aster docker image


    :param inp_path: Path to input file folder(s)
    :param cpus: Number of CPUs to run the analysis on. Default is 2.
    :param gpus: Number of GPUs to run the analysis on. Default is none.
    :param run_ext: If False the process will wait for the abaqus analysis to finish. Default is False
    :param metadata: Dictionary containing various metadata relevant for the analysis
    :param execute: Automatically starts Abaqus analysis. Default is True
    :param return_bat_str:
    :param exit_on_complete:
    """

    ca = CodeAsterAnalysis(
        inp_path,
        cpus=cpus,
        run_ext=run_ext,
        metadata=metadata,
        execute=execute,
    )
    ca.run(exit_on_complete=exit_on_complete)


class CodeAsterAnalysis(LocalExecute):
    def __init__(self, inp_path, cpus=2, execute=True, metadata=None, local_execute=True, run_ext=True):
        """
        Code Aster Analysis

        Local Installation from:
        https://bitbucket.org/siavelis/codeaster-windows-src/downloads/code-aster_v2019_std-win64.zip

        :param inp_path:
        :param cpus:
        """
        super(CodeAsterAnalysis, self).__init__(
            inp_path,
            cpus,
            gpus=None,
            run_ext=run_ext,
            metadata=metadata,
            excute_locally=local_execute,
            auto_execute=execute,
        )

    def _run_docker(self):
        raise NotImplementedError()
        # try:
        #     import docker
        # except ModuleNotFoundError as e:
        #     raise ModuleNotFoundError(
        #         "To use docker functionality you will need to install docker first.\n"
        #         'Use "pip install docker"\n\n'
        #         f'Original error message: "{e}"'
        #     )
        # cpu_count = 2
        #
        # calc_file = pathlib.Path(calc_file)
        # start_time = time.time()
        # environment = dict()
        # client.images.pull("quay.io/tianyikillua/code_aster", "latest")
        # container = client.containers.run(
        #     "quay.io/tianyikillua/code_aster",
        #     f"/home/aster/aster/bin/as_run {calc_file}",
        #     detach=True,
        #     working_dir="/home/calc/",
        #     environment=environment,
        #     cpu_count=cpu_count,
        #     volumes={work_dir: {"bind": "/home/calc/", "mode": "rw"}},
        # )
        #
        # for line in container.logs(stream=True):
        #     print(line.strip().decode("utf-8"))
        # end_time = time.time()
        # res_str = f"Analysis time {end_time - start_time:.2f}s"
        # print(res_str)

    def run(self, exit_on_complete=True):
        if Settings.use_docker_execute is False:
            try:
                exe_path = get_exe_path("code_aster")
            except FileNotFoundError as e:
                logging.error(e)
                return

            self._run_local(f'"{exe_path}" {self.analysis_name}.export', exit_on_complete=exit_on_complete)
        else:
            self._run_docker()



# coding=utf-8
import numpy as np


class Metal:
    """
    Base object for all metals
    """

    def __init__(
        self,
        E,
        rho,
        sig_y,
        sig_u,
        v,
        zeta,
        alpha,
        plasticitymodel,
        eps_p=None,
        sig_p=None,
        units="m",
    ):
        self._E = E
        self._rho = rho
        self._sig_y = sig_y
        self._sig_u = sig_u
        self._v = v
        self.zeta = zeta
        self._alpha = alpha
        self._plasticitymodel = plasticitymodel
        self._eps_p = eps_p
        self._sig_p = sig_p
        self._units = units

    def __delattr__(self, item):
        raise AttributeError("Deletion of base material object properties is not allowed!")

    # def __setattr__(self, key, value):
    #     if key in self.__dict__:
    #         raise AttributeError("Can't set attribute {!r} on base material object".format(key))
    #     self.__dict__[key] = value

    def __repr__(self):
        return f"Metal(E:{self.E}, rho:{self.rho}, Sigy: {self.sig_y}, Plasticity Model: {self.plasticity_model})"

    @property
    def E(self):
        """

        :return: Young's Modulus
        """
        return self._E

    @E.setter
    def E(self, value):
        if value < 0:
            raise ValueError("The Young's Modulus must be a positive number")
        self._E = value

    @property
    def sig_y(self):
        """
        Yield stress

        :return:
        """
        return self._sig_y

    @sig_y.setter
    def sig_y(self, value):
        if value < 0:
            raise ValueError("Yield Stress must be a positive number")
        self._sig_y = value

    @property
    def sig_u(self):
        """
        :return: Ultimate yield stress
        """
        return self._sig_u

    @property
    def rho(self):
        """
        :return: Density
        """
        return self._rho

    @property
    def v(self):
        """
        Poisson Ratio

        :return:
        """
        return self._v

    @property
    def alpha(self):
        """
        Thermal Expansion coefficient

        :return:
        """
        return self._alpha

    @property
    def zeta(self):
        """
        Material damping coefficient

        :return:
        """
        return self._zeta

    @zeta.setter
    def zeta(self, value):
        if value is None or value < 0.0:
            raise ValueError("Zeta cannot be None or below zero")
        self._zeta = value

    @property
    def plasticity_model(self):
        """

        :return: Constitutive Equation for plasticity
        """
        return self._plasticitymodel

    @property
    def eps_p(self):
        if self._eps_p is not None:
            return self._eps_p
        elif self._plasticitymodel is not None:
            return self._plasticitymodel.eps_p
        else:
            return None

    @property
    def sig_p(self):
        if self._sig_p is not None:
            return self._sig_p
        elif self._plasticitymodel is not None:
            return self._plasticitymodel.sig_p
        else:
            return None

    @property
    def E_therm(self):
        return None

    @property
    def sigy_therm(self):
        return None

    @property
    def kappa(self):
        return None

    @property
    def cp(self):
        return None

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if self._units == "m" and value == "mm":
            self._E *= 1e-6
            if self._sig_y is not None:
                self._sig_y *= 1e-6
            self._rho *= 1e-9
            self._units = value
        elif self._units == "mm" and value == "m":
            self._E *= 1e6
            if self._sig_y is not None:
                self._sig_y *= 1e6
            self._rho *= 1e9
            self._units = value


class DNVGL16PBase:
    def __init__(self, q_prop, q_yield_1, q_yield_2, ep_y1, ep_y2, n, K):
        self.q_prop = q_prop
        self.q_yield_1 = q_yield_1
        self.q_yield_2 = q_yield_2
        self.ep_y1 = ep_y1
        self.ep_y2 = ep_y2
        self.n = n
        self.K = K


class DnvGl16Mat:
    """
    This function returns a carbon steel Material object based on the formulas in DNVGL RP C208 (september 2016)


    :param t: Thickness of the material
    :param grade: Material Grade 'S355' or 'S420'
    :param mat_def: Material Definition
    :param eps_max: Maximum epsilon
    :param data_points: Number of datapoints
    :return: tuple of (eps_p, sig_p)
    """

    _params = {
        "Low": {
            "t_16": {
                "q_prop": {"S355": 320.0e6, "S420": 378.7e6},
                "q_yield_1": {"S355": 357.0e6, "S420": 422.5e6},
                "q_yield_2": {"S355": 366.1e6, "S420": 426.3e6},
                "ep_y1": {"S355": 0.004, "S420": 0.004},
                "ep_y2": {"S355": 0.015, "S420": 0.012},
                "n": {"S355": 0.166, "S420": 0.14},
                "K": {"S355": 740e6, "S420": 738e6},
            },
            "16_t_40": {
                "q_prop": {"S355": 311.0e6, "S420": 360.6e6},
                "q_yield_1": {"S355": 346.9e6, "S420": 402.4e6},
                "q_yield_2": {"S355": 355.9e6, "S420": 406.0e6},
                "ep_y1": {"S355": 0.004, "S420": 0.004},
                "ep_y2": {"S355": 0.015, "S420": 0.012},
                "n": {"S355": 0.166, "S420": 0.14},
                "K": {"S355": 740e6, "S420": 703e6},
            },
            "40_t_63": {
                "q_prop": {"S355": 301.9e6, "S420": 351.6e6},
                "q_yield_1": {"S355": 336.9e6, "S420": 392.3e6},
                "q_yield_2": {"S355": 345.7e6, "S420": 395.9e6},
                "ep_y1": {"S355": 0.004, "S420": 0.004},
                "ep_y2": {"S355": 0.015, "S420": 0.012},
                "n": {"S355": 0.166, "S420": 0.14},
                "K": {"S355": 725e6, "S420": 686e6},
            },
            "63_t_100": {
                "q_prop": {"S355": 284e6},
                "q_yield_1": {"S355": 316.7e6},
                "q_yield_2": {"S355": 323.8e6},
                "ep_y1": {"S355": 0.004},
                "ep_y2": {"S355": 0.015},
                "n": {"S355": 0.166},
                "K": {"S355": 725e6},
            },
        },
        "Mean": {
            "t_16": {
                "q_prop": {"S355": 384.0e6, "S420": 435.5e6},
                "q_yield_1": {"S355": 428.4e6, "S420": 482.9e6},
                "q_yield_2": {"S355": 439.3e6, "S420": 487.2e6},
                "ep_y1": {"S355": 0.004, "S420": 0.004},
                "ep_y2": {"S355": 0.015, "S420": 0.011928571},
                "n": {"S355": 0.166, "S420": 0.14},
                "K": {"S355": 900e6, "S420": 738e6},
            },
            "16_t_40": {
                "q_prop": {"S355": 357.7e6, "S420": 432.7e6},
                "q_yield_1": {"S355": 398.9e6, "S420": 482.9e6},
                "q_yield_2": {"S355": 409.3e6, "S420": 487.2e6},
                "ep_y1": {"S355": 0.004, "S420": 0.004},
                "ep_y2": {"S355": 0.015, "S420": 0.011929},
                "n": {"S355": 0.166, "S420": 0.14},
                "K": {"S355": 850e6, "S420": 703e6},
            },
            "40_t_63": {
                "q_prop": {"S355": 332.1e6, "S420": 421.9e6},
                "q_yield_1": {"S355": 370.6e6, "S420": 470.8e6},
                "q_yield_2": {"S355": 380.3e6, "S420": 475.1e6},
                "ep_y1": {"S355": 0.004, "S420": 0.004},
                "ep_y2": {"S355": 0.015, "S420": 0.011929},
                "n": {"S355": 0.166, "S420": 0.14},
                "K": {"S355": 800e6, "S420": 686e6},
            },
            "63_t_100": {
                "q_prop": {"S355": 312.4e6},
                "q_yield_1": {"S355": 348.4e6},
                "q_yield_2": {"S355": 350.6e6},
                "ep_y1": {"S355": 0.004},
                "ep_y2": {"S355": 0.015},
                "n": {"S355": 0.166},
                "K": {"S355": 800e6},
            },
        },
    }

    @staticmethod
    def dnv_thick_select(thick, grade):
        if thick < 0.016:
            mat_str = "t_16"
        elif 0.016 <= thick < 0.04:
            mat_str = "16_t_40"
        elif 0.04 <= thick < 0.063:
            mat_str = "40_t_63"
        else:
            if grade == "S420":
                mat_str = "40_t_63"
            else:
                mat_str = "63_t_100"
        return mat_str

    def __init__(self, t, grade, mat_def="Low", eps_max=0.3, data_points=200):
        self._mat_def = mat_def
        self._grade = grade
        self._t = t
        thick_str = self.dnv_thick_select(t, grade)
        sig_prop = float(self._params[mat_def][thick_str]["q_prop"][grade])
        sig_yield_1 = float(self._params[mat_def][thick_str]["q_yield_1"][grade])
        sig_yield_2 = float(self._params[mat_def][thick_str]["q_yield_2"][grade])
        ep_y1 = float(self._params[mat_def][thick_str]["ep_y1"][grade])
        ep_y2 = float(self._params[mat_def][thick_str]["ep_y2"][grade])
        n = float(self._params[mat_def][thick_str]["n"][grade])
        K = float(self._params[mat_def][thick_str]["K"][grade])

        eps = np.linspace(ep_y2, eps_max, num=data_points)

        ep = list()
        ep.append(0.0)
        ep.append(ep_y1)
        ep.append(ep_y2)
        sig = list()
        sig.append(sig_prop)
        sig.append(sig_yield_1)
        sig.append(sig_yield_2)

        init_val = 0
        for e in eps:
            if init_val == 0:
                init_val = 1
            else:
                ep.append(e)
                sig.append(K * (e + (sig_yield_2 / K) ** (1.0 / n) - ep_y2) ** n)

        self._eps_p = ep
        self._sig_p = sig

    def __repr__(self):
        return f"Dnvgl16Mat({self.nl_name})"

    @property
    def eps_p(self):
        return self._eps_p

    @property
    def sig_p(self):
        return self._sig_p

    @property
    def nl_name(self):
        mat_prefix = "RP"
        mat_def_str = "L" if self._mat_def == "Low" else "M"
        return mat_prefix + "_" + self._grade + mat_def_str + "_" + self.dnv_thick_select(self._t, self._grade)

    @property
    def metadata(self):
        return dict(description="Carbon Steel nonlinear material based on DNVGL-RP-C208 (Sept. 2019)")


class CarbonSteel(Metal):
    GRADES = dict(
        S355=dict(name="S355", sigy=355e6, sigu=355e6),
        S420=dict(name="S420", sigy=420e6, sigu=420e6),
    )
    EC3_TEMP = [20, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 1100, 1200]
    EC3_E_RED = [
        1.0,
        1.0,
        0.9,
        0.8,
        0.7,
        0.6,
        0.31,
        0.13,
        0.09,
        0.0675,
        0.045,
        0.0225,
        0.0,
    ]
    EC3_S_RED = [1.0, 1.0, 1.0, 1.0, 1.0, 0.78, 0.47, 0.23, 0.11, 0.06, 0.04, 0.02, 0.0]

    """

    :param grade: Material Grade
    :param plasticity_model: Plasticity model e.g. CarbonSteel
    :param E: Young's Modulus
    :param rho: Material Density
    :param v: Poisson Ratio
    :param zeta: Material damping coefficient
    :param alpha: Thermal Expansion coefficient
    :param sig_y: Yield stress
    :param sig_u: Ultimate stress
    :param eps_p: Plastic strain
    :param sig_p: Plastic stress
    :param temp_range: Temperature range
    :param units: Definition of length unit. Default is meter 'm'. Alternative is millimeter 'mm'.
    """

    def __init__(
        self,
        grade="S355",
        plasticity_model=DnvGl16Mat(t=0.01, grade="S355", mat_def="Low", eps_max=0.3, data_points=200),
        E=2.1e11,
        rho=7850,
        v=0.3,
        zeta=1.15,
        alpha=1.2e-5,
        sig_y=None,
        sig_u=None,
        eps_p=None,
        sig_p=None,
        temp_range=None,
        units="m",
    ):
        self._grade = grade
        sig_y = sig_y if sig_y is not None else CarbonSteel.GRADES[grade]["sigy"]
        sig_u = sig_u if sig_u is not None else CarbonSteel.GRADES[grade]["sigu"]

        Metal.__init__(
            self,
            E,
            rho,
            sig_y,
            sig_u,
            v,
            zeta,
            alpha,
            plasticity_model,
            sig_p=sig_p,
            eps_p=eps_p,
            units=units,
        )
        # Manually override variables

        self._temp_range = np.arange(20, 1210, 5) if temp_range is None else temp_range

    def __repr__(self):
        return (
            f"CarbonSteel(E:{self.E:.3E}, sig_y:{self.sig_y:.3E}, rho:{self.rho:.3E},"
            f" plasticity_model:{self.plasticity_model})"
        )

    @property
    def grade(self):
        return self._grade

    @property
    def temp_range(self):
        return self._temp_range

    @property
    def E_therm(self):
        E_red_fac = np.interp(self._temp_range, CarbonSteel.EC3_TEMP, CarbonSteel.EC3_E_RED)
        return [self._E * x for x in E_red_fac]

    @property
    def sigy_therm(self):
        sig_red_fac = np.interp(self._temp_range, CarbonSteel.EC3_TEMP, CarbonSteel.EC3_S_RED)
        return [self._sig_y * x for x in sig_red_fac]

    @property
    def kappa(self):
        """
        Thermal conductivity

        watts per meter-kelvin W/(mK)

        :return:
        """

        phase1_end = 780
        phase1_arr = [self._temp_range[x] for x in np.where(self._temp_range <= phase1_end)]
        phase2_arr = [self._temp_range[x] for x in np.where(self._temp_range > phase1_end)]
        phase1 = [54 - t * 3.33 * 0.01 for t in phase1_arr[0]]
        phase2 = [27.3 for x in range(phase2_arr[0].shape[0])]
        return phase1 + phase2

    @property
    def cp(self):
        """
        Specific Heat

        joule per kelvin and kilogram J/(K kg)

        :return:
        """

        phase1_end = 600
        phase2_end = 735
        phase3_end = 900
        phase1_arr = [self._temp_range[x] for x in np.where(self._temp_range <= phase1_end)]
        phase2_arr = [
            self._temp_range[x]
            for x in np.where(np.logical_and(self._temp_range > phase1_end, self._temp_range <= phase2_end))
        ]
        phase3_arr = [
            self._temp_range[x]
            for x in np.where(np.logical_and(self._temp_range > phase2_end, self._temp_range <= phase3_end))
        ]
        phase4_arr = [self._temp_range[x] for x in np.where(self._temp_range > phase3_end)]
        phase1 = [425 + 7.73 * 1e-1 * t - 1.69 * 1e-3 * (t ** 2) + 2.22 * 1e-6 * t ** 3 for t in phase1_arr[0]]
        phase2 = [666 + 13002 / (738 - t) for t in phase2_arr[0]]
        phase3 = [545 + 17820 / (t - 731) for t in phase3_arr[0]]
        phase4 = [650 for x in range(phase4_arr[0].shape[0])]
        return phase1 + phase2 + phase3 + phase4


class Aluminium(Metal):
    def __init__(self):
        raise NotImplementedError("The aluminium material model is not yet implemented")



import json
import pathlib
import re
from collections import OrderedDict

import ada.core.utils

from . import SectionCat


def profile_db_collect(sec_type, dim, units="m"):
    """
    Tale in section type and dimensions and return a section object based on values in a
    profile db xml document. The elements in the xml doc is in units meters.

    :param sec_type:
    :param dim:
    :param units:
    :type sec_type: str
    :type dim: str
    :type units: str
    """
    from ada import Section

    if units == "mm":
        scale_factor = 1000
    elif units == "m":
        scale_factor = 1.0
    else:
        raise ValueError(f'Units "{units}" is not supported')

    if sec_type == "IP":
        sec_type = "IPE"
    dir_path = pathlib.Path(__file__).resolve().parent

    with open(dir_path / "resources" / "ProfileDB.json") as data_file:
        profile_db_main = json.load(data_file, object_pairs_hook=OrderedDict)

    profile_db = profile_db_main["ProfileDB"]
    sec_dim = [int(x) for x in dim.split("x")]
    sec_name_alt1 = sec_type + str(dim)
    if "x" in str(dim):
        sec_name_alt2 = sec_type + str(sec_dim[0]) + "x" + str(sec_dim[1])
    else:
        sec_name_alt2 = "Unknown"

    if sec_type in profile_db.keys():
        if sec_name_alt1 in profile_db[sec_type]:
            sec_name = sec_name_alt1
        elif sec_name_alt2 in profile_db[sec_type]:
            sec_name = sec_name_alt2
        else:
            return None
        h = float(profile_db[sec_type][sec_name]["Height"]) * scale_factor
        w_top = float(profile_db[sec_type][sec_name]["Width"]) * scale_factor
        w_btn = float(profile_db[sec_type][sec_name]["Width"]) * scale_factor
        t_w = float(profile_db[sec_type][sec_name]["t_w"]) * scale_factor
        t_fbtn = float(profile_db[sec_type][sec_name]["t_f"]) * scale_factor
        t_ftop = float(profile_db[sec_type][sec_name]["t_f"]) * scale_factor
        return Section(
            sec_name,
            sec_type=sec_type,
            h=h,
            w_top=w_top,
            w_btn=w_btn,
            t_w=t_w,
            t_fbtn=t_fbtn,
            t_ftop=t_ftop,
            sec_str=sec_name,
            metadata=dict(cad_str=sec_name),
            units=units,
        )
    else:
        return None


def interpret_section_str(in_str, s=0.001, units="m"):
    """

    :param in_str:
    :param s: Scale factor
    :param units: The desired units after applied scale factor
    :return: Two section (to account for potential beam tapering)
    """
    from ada import Section

    rdoff = ada.core.utils.roundoff

    digit = r"\d{0,5}\.?\d{0,5}|\d{0,5}|\d{0,5}\/\d{0,5}"
    flex = r"?:\.|[A-Z]|"
    re_in = re.IGNORECASE | re.DOTALL
    for box in SectionCat.box:
        res = re.search(
            r"({box})({flex})({digit})x({digit})x({digit})x({digit})".format(box=box, flex=flex, digit=digit),
            in_str,
            re_in,
        )

        if res is not None:
            h = [rdoff(float(x) * s) for x in res.group(2).split("/")]
            width = [rdoff(float(x) * s) for x in res.group(3).split("/")]
            tw = [rdoff(float(x) * s) for x in res.group(4).split("/")]
            tf = [rdoff(float(x) * s) for x in res.group(5).split("/")]
            sec = Section(
                in_str,
                h=h[0],
                sec_type=box,
                w_btn=width[0],
                w_top=width[0],
                t_fbtn=tf[0],
                t_ftop=tf[0],
                t_w=tw[0],
                metadata=dict(cad_str=in_str),
                units=units,
            )
            tap = Section(
                in_str + "_e",
                h=h[-1],
                sec_type=box,
                w_btn=width[-1],
                w_top=width[-1],
                t_fbtn=tf[-1],
                t_ftop=tf[-1],
                t_w=tw[-1],
                metadata=dict(cad_str=in_str),
                units=units,
            )
            return sec, tap

    for shs in SectionCat.shs:
        res = re.search(
            r"({hol})({flex})({digit})x({digit})".format(hol=shs, flex=flex, digit=digit),
            in_str,
            re_in,
        )

        if res is not None:
            h = [rdoff(float(x) * s) for x in res.group(2).split("/")]
            width = [rdoff(float(x) * s) for x in res.group(3).split("/")]
            sec = Section(
                in_str,
                h=h[0],
                sec_type=shs,
                w_btn=h[0],
                w_top=h[0],
                t_fbtn=width[0],
                t_ftop=width[0],
                t_w=width[0],
                metadata=dict(cad_str=in_str),
                units=units,
            )
            tap = Section(
                in_str + "_e",
                h=h[-1],
                sec_type=shs,
                w_btn=h[-1],
                w_top=h[-1],
                t_fbtn=width[-1],
                t_ftop=width[-1],
                t_w=width[-1],
                metadata=dict(cad_str=in_str),
                units=units,
            )
            return sec, tap

    for rhs in SectionCat.rhs:
        res = re.search(
            r"({rhs})({flex})({digit})x({digit})x({digit})".format(rhs=rhs, flex=flex, digit=digit),
            in_str,
            re_in,
        )

        if res is not None:
            h = [rdoff(float(x) * s) for x in res.group(2).split("/")]
            width = [rdoff(float(x) * s) for x in res.group(3).split("/")]
            tw = [rdoff(float(x) * s) for x in res.group(4).split("/")]
            sec = Section(
                in_str,
                h=h[0],
                sec_type=rhs,
                w_btn=width[0],
                w_top=width[0],
                t_fbtn=tw[0],
                t_ftop=tw[0],
                t_w=tw[0],
                metadata=dict(cad_str=in_str),
                units=units,
            )
            tap = Section(
                in_str + "_e",
                h=h[-1],
                sec_type=rhs,
                w_btn=width[-1],
                w_top=width[-1],
                t_fbtn=tw[-1],
                t_ftop=tw[-1],
                t_w=tw[-1],
                metadata=dict(cad_str=in_str),
                units=units,
            )
            return sec, tap

    for ig in SectionCat.igirders:
        res = re.search(
            "({ig})({flex})({digit})x({digit})x({digit})x({digit})".format(ig=ig, flex=flex, digit=digit),
            in_str,
            re_in,
        )
        if res is not None:
            h = [rdoff(float(x) * s) for x in res.group(2).split("/")]
            wt = [rdoff(float(x) * s) for x in res.group(3).split("/")]
            tw = [rdoff(float(x) * s) for x in res.group(4).split("/")]
            tf = [rdoff(float(x) * s) for x in res.group(5).split("/")]
            sec = Section(
                in_str,
                h=h[0],
                sec_type=ig,
                w_btn=wt[0],
                w_top=wt[0],
                t_fbtn=tf[0],
                t_ftop=tf[0],
                t_w=tw[0],
                metadata=dict(cad_str=in_str),
                units=units,
            )
            tap = Section(
                in_str + "_e",
                h=h[-1],
                sec_type=ig,
                w_btn=wt[-1],
                w_top=wt[-1],
                t_fbtn=tf[-1],
                t_ftop=tf[-1],
                t_w=tw[-1],
                metadata=dict(cad_str=in_str),
                units=units,
            )
            return sec, tap

    if "pipe" not in in_str.lower():
        for ipe in SectionCat.iprofiles:
            res = re.search("({ipe})({digit})".format(ipe=ipe, digit=digit), in_str, re_in)
            if res is not None:
                sec = profile_db_collect(ipe, res.group(2), units=units)
                return sec, sec

    for tub in SectionCat.tubular:
        res = re.search("({tub})({digit})x({digit})".format(tub=tub, digit=digit), in_str, re_in)
        if res is not None:
            fac = 0.5 if tub == "OD" else 1.0
            r = [rdoff(float(x) * s * fac) for x in res.group(2).split("/")]
            wt = [rdoff(float(x) * s) for x in res.group(3).split("/")]
            sec = Section(
                in_str,
                sec_type=tub,
                r=r[0],
                wt=wt[0],
                metadata=dict(cad_str=in_str),
                units=units,
            )
            tap = Section(
                in_str + "_e",
                sec_type=tub,
                r=r[-1],
                wt=wt[-1],
                metadata=dict(cad_str=in_str),
                units=units,
            )
            return sec, tap

    for ang in SectionCat.angular:
        res = re.search("({ang})({digit})x({digit})".format(ang=ang, digit=digit), in_str, re_in)
        if res is not None:
            sec = profile_db_collect(ang, "{}x{}".format(res.group(2), res.group(3)), units=units)
            return sec, sec

    for circ in SectionCat.circular:
        res = re.search("{circ}(.*?)$".format(circ=circ), in_str, re_in)
        if res is not None:
            sec = Section(
                in_str,
                sec_type=circ,
                r=rdoff(float(res.group(1)) * s),
                metadata=dict(cad_str=in_str),
                units=units,
            )
            return sec, sec

    raise ValueError(f'Unable to interpret section str "{in_str}"')



import logging
from itertools import groupby
from operator import attrgetter

import numpy as np

from ada.config import Settings as _Settings
from ada.fem import ElemShapes

from ..utils import _folder_prep, get_fem_model_from_assembly
from .common import abaqus_to_med_type


def to_fem(
    assembly,
    name,
    scratch_dir=None,
    description=None,
    execute=False,
    run_ext=False,
    cpus=2,
    gpus=None,
    overwrite=False,
    exit_on_complete=True,
):
    """
    Write Code_Aster .med, .export and .comm file from Assembly data


    MED writer modified and based on meshio implementation..

    :param assembly:
    :param name:
    :param scratch_dir:
    :param description:
    :param execute:
    :param run_ext:
    :param cpus:
    :param gpus:
    :param overwrite:
    :param exit_on_complete:
    """

    analysis_dir = _folder_prep(scratch_dir, name, overwrite)

    if "info" not in assembly.metadata:
        assembly.metadata["info"] = dict(description="")

    assembly.metadata["info"]["description"] = description

    p = get_fem_model_from_assembly(assembly)

    # TODO: Implement support for multiple parts. Need to understand how submeshes in Salome and Code Aster works.
    # for p in filter(lambda x: len(x.fem.elements) != 0, assembly.get_all_parts_in_assembly(True)):
    write_to_med(name, p, analysis_dir)

    with open((analysis_dir / name).with_suffix(".export"), "w") as f:
        f.write(write_export_file(name, cpus))

    with open((analysis_dir / name).with_suffix(".comm"), "w") as f:
        f.write(write_to_comm(assembly, p))

    print(f'Created a Code_Aster input deck at "{analysis_dir}"')

    from .execute import run_code_aster

    run_code_aster(
        (analysis_dir / name).with_suffix(".export"),
        cpus=cpus,
        gpus=gpus,
        run_ext=run_ext,
        metadata=assembly.metadata,
        execute=execute,
        exit_on_complete=exit_on_complete,
    )


# COMM File input
def write_to_comm(assembly, part):
    """
    Create COMM file input str

    :param assembly:
    :type assembly: ada.Assembly
    :param part:
    :type part: ada.Part
    :return: COMM input file str
    """

    materials_str = "\n".join(list(map(write_material, part.materials)))
    sections_str = write_sections(part.fem.sections)
    bc_str = "\n".join(list(map(write_boundary_condition, assembly.fem.bcs)))
    step_str = "\n".join([write_step(s, part) for s in assembly.fem.steps])

    comm_str = f"""#
#   COMM file created by ADA (Assembly for Design and Analysis)
#

# Units: N, m

DEBUT(LANG="EN")

mesh = LIRE_MAILLAGE(FORMAT="MED", UNITE=20)

model = AFFE_MODELE(
    AFFE=_F(MODELISATION=("DKT",), PHENOMENE="MECANIQUE", TOUT="OUI"), MAILLAGE=mesh
)

# Materials
{materials_str}

# Sections
{sections_str}

# Boundary Conditions
{bc_str}

# Step Information
{step_str}

FIN()
"""

    return comm_str


def write_material(material):
    """

    :param material:
    :type material: ada.Material
    :return: Code Aster COMM Input string
    """
    from ada.core.utils import NewLine

    # Bi-linear hardening ECRO_LINE=_F(D_SIGM_EPSI=2.0e06, SY=2.35e06,)

    model = material.model
    nl = NewLine(3, suffix="	")
    nl_mat = "nl_mat=(	\n	"

    if model.plasticity_model is not None:
        eps = [e for e in model.eps_p]
        eps[0] = 1e-5  # Epsilon index=0 cannot be zero
        nl_mat += "".join([f"{e:.4E},{s:.4E}," + next(nl) for e, s in zip(eps, model.sig_p)]) + ")"
    else:
        nl_mat += ",)"
        logging.error(f"No plasticity is defined for material {material.name}")

    hardening_model_str = """Traction=DEFI_FONCTION(
    NOM_PARA='EPSI', NOM_RESU='SIGM', VALE=nl_mat, INTERPOL='LIN', PROL_DROITE='LINEAIRE', PROL_GAUCHE='CONSTANT'
)"""
    return f"""{nl_mat}

{hardening_model_str}

{material.name} = DEFI_MATERIAU(
    ELAS=_F(E={model.E}, NU={model.v}, RHO={model.rho}), TRACTION=_F(SIGM=Traction,),
)
"""


def write_sections(fem_sections):
    """

    :param fem_sections:
    :type fem_sections: ada.fem.containers.FemSections
    :return:
    """
    mat_assign_str = ""
    shell_sections_str = ""

    if len(fem_sections.shells) > 0:
        mat_assign_str, shell_sections_str = [
            "".join(x) for x in zip(*list(map(write_shell_section, fem_sections.shells)))
        ]

    if len(fem_sections.beams) > 0:
        logging.error("Beam section export to Code Aster is not yet supported")

    if len(fem_sections.solids) > 0:
        logging.error("Solid section export to Code Aster is not yet supported")

    beam_sections_str = ""
    solid_sections_str = ""
    return f"""
material = AFFE_MATERIAU(MODELE=model, AFFE=(
{mat_assign_str}
    )
)


# Shell elements:
#   EPAIS: thickness
#   VECTEUR: a direction of reference in the tangent plan

element = AFFE_CARA_ELEM(
        MODELE=model,
        COQUE=(
{shell_sections_str}{beam_sections_str}{solid_sections_str}
        ),
    )
"""


def write_shell_section(fem_sec):
    """


    :param fem_sec:
    :type fem_sec: ada.fem.FemSection
    :return:
    """
    mat_name = fem_sec.material.name
    sec_name = fem_sec.elset.name
    #
    local_vec = str(tuple(fem_sec.local_y))
    mat_ = f'		_F(MATER=({mat_name},), GROUP_MA="{sec_name}"),\n'
    sec_str = f"""            _F(
                GROUP_MA=("{sec_name}"),
                EPAIS={fem_sec.thickness},
                VECTEUR={local_vec},
            ),
"""
    return mat_, sec_str


def write_boundary_condition(bc):
    """

    :param bc:
    :type bc: ada.fem.Bc
    :return:
    :rtype: str
    """
    set_name = bc.fem_set.name
    bc_str = ""
    for i, n in enumerate(["DX", "DY", "DZ", "DRX", "DRY", "DRZ"], start=1):
        if i in bc.dofs:
            bc_str += f"{n}=0, "
    dofs_str = f"""dofs = dict(
    GROUP_NO="{set_name}",
    {bc_str}
)\n"""

    return (
        dofs_str
        + f"""{bc.name}_bc = AFFE_CHAR_MECA(
    MODELE=model, DDL_IMPO=_F(**dofs)
)"""
    )


def write_load(load):
    """

    :param load:
    :type load: ada.fem.Load
    :return:
    :rtype: str
    """
    if load.type == "gravity":
        return f"""{load.name} = AFFE_CHAR_MECA(
    MODELE=model, PESANTEUR=_F(DIRECTION=(0.0, 0.0, -1.0), GRAVITE={-load.magnitude})
)"""
    else:
        raise NotImplementedError(f'Load type "{load.type}"')


def write_step(step, part):
    """

    :param step:
    :type step: ada.fem.Step
    :param part:
    :type part: ada.Part
    :return:
    :rtype: str
    """

    load_str = "\n".join(list(map(write_load, step.loads)))
    load = step.loads[0]
    bc = part.get_assembly().fem.bcs[0]
    if step.type == "static":
        return f"""
{load_str}

timeReel = DEFI_LIST_REEL(DEBUT=0.0, INTERVALLE=_F(JUSQU_A=1.0, NOMBRE=10))
timeInst = DEFI_LIST_INST(METHODE="AUTO", DEFI_LIST=_F(LIST_INST=timeReel))
rampFunc = DEFI_FONCTION(NOM_PARA="INST", VALE=(0.0, 0.0, 1.0, 1.0))

result = STAT_NON_LINE(
    MODELE=model,
    CHAM_MATER=material,
    CARA_ELEM=element,
    COMPORTEMENT=(_F(DEFORMATION="PETIT", RELATION="VMIS_ISOT_TRAC", TOUT="OUI")),
    CONVERGENCE=_F(ARRET="OUI", ITER_GLOB_MAXI=8,),
    EXCIT=(_F(CHARGE={bc.name}_bc), _F(CHARGE={load.name}, FONC_MULT=rampFunc)),
    INCREMENT=_F(LIST_INST=timeInst),
    ARCHIVAGE=_F(LIST_INST=timeReel),
)

result = CALC_CHAMP(
    reuse=result, RESULTAT=result,
    CONTRAINTE=("EFGE_ELNO", "EFGE_NOEU", "SIGM_ELNO"),
    DEFORMATION=("EPSI_ELNO", "EPSP_ELNO"),
)

stress = POST_CHAMP(
    EXTR_COQUE=_F(
        NIVE_COUCHE='MOY',
    NOM_CHAM=('SIGM_ELNO', ),
    NUME_COUCHE=1),
    RESULTAT=result
)

stress = CALC_CHAMP(
    reuse=stress,
    CONTRAINTE=('SIGM_NOEU', ),
    RESULTAT=stress
)

strain = POST_CHAMP(
    EXTR_COQUE=_F(
        NIVE_COUCHE='MOY',
    NOM_CHAM=('EPSI_ELNO', ),
    NUME_COUCHE=1),
    RESULTAT=result
)

strainP = POST_CHAMP(
    EXTR_COQUE=_F(
        NIVE_COUCHE='MOY',
    NOM_CHAM=('EPSP_ELNO', ),
    NUME_COUCHE=1),
    RESULTAT=result
)

IMPR_RESU(
    RESU=(
        _F(
            NOM_CHAM=("DEPL", "EFGE_ELNO", "EFGE_NOEU"),
            NOM_CHAM_MED=("DISP", "GEN_FORCES_ELEM", "GEN_FORCES_NODES"),
            RESULTAT=result,
        ),
        _F(
            NOM_CHAM=("SIGM_ELNO", "SIGM_NOEU"),
            NOM_CHAM_MED=("STRESSES_ELEM", "STRESSES_NODES"),
            RESULTAT=stress,
        ),
        _F(
            NOM_CHAM=("EPSI_ELNO",),
            NOM_CHAM_MED=("STRAINS_ELEM",),
            RESULTAT=strain,
        ),
        _F(
            NOM_CHAM=("EPSP_ELNO",),
            NOM_CHAM_MED=("PLASTIC_STRAINS_ELEM",),
            RESULTAT=strainP,
        ),
    ),
    UNITE=80,
)"""
    else:
        raise NotImplementedError(f'Step stype "{step.type}" is not yet supported')


# EXPORT file input
def write_export_file(name, cpus):
    """

    :param name:
    :param cpus:
    :return:
    """
    export_str = rf"""P actions make_etude
P memory_limit 1274
P time_limit 900
P version stable
P mpi_nbcpu 1
P mode interactif
P ncpus {cpus}
F comm {name}.comm D 1
F mmed {name}.med D 20
F mess {name}.mess R 6
F rmed {name}.rmed R 80"""

    return export_str


# MED file input
def write_to_med(name, part, analysis_dir):
    """
    Custom Method for writing a part directly based on meshio

    :param name:
    :param part:
    :param analysis_dir:
    :type part: ada.Part
    :return:
    """
    import pathlib

    import h5py

    analysis_dir = pathlib.Path(analysis_dir)
    filename = (analysis_dir / name).with_suffix(".med")

    f = h5py.File(filename, "w")
    mesh_name = name if name is not None else part.fem.name
    # Strangely the version must be 3.0.x
    # Any version >= 3.1.0 will NOT work with SALOME 8.3
    info = f.create_group("INFOS_GENERALES")
    info.attrs.create("MAJ", 3)
    info.attrs.create("MIN", 0)
    info.attrs.create("REL", 0)

    time_step = _write_mesh_presets(f, mesh_name)

    profile = "MED_NO_PROFILE_INTERNAL"

    # Node and Element sets (familles in French)
    fas = f.create_group("FAS")
    families = fas.create_group(mesh_name)
    family_zero = families.create_group("FAMILLE_ZERO")  # must be defined in any case
    family_zero.attrs.create("NUM", 0)

    # Make sure that all member references are updated (TODO: Evaluate if this can be avoided using a smarter algorithm)
    part.fem.sets.add_references()

    # Nodes and node sets
    _write_nodes(part, time_step, profile, families)

    # Elements (mailles in French) and element sets
    _write_elements(part, time_step, profile, families)

    f.close()


def _write_nodes(part, time_step, profile, families):
    """

    TODO: Go through each data group and set in HDF5 file and make sure that it writes what was read 1:1.
        Use cylinder.med as a benchmark.

    Add the following datasets ['COO', 'FAM', 'NUM'] to the 'NOE' group

    :param part:
    :param time_step:
    :param profile:
    :return:
    """
    points = np.zeros((int(part.fem.nodes.max_nid), 3))

    def pmap(n):
        points[int(n.id - 1)] = n.p

    list(map(pmap, part.fem.nodes))

    # Try this
    if _Settings.ca_experimental_id_numbering is True:
        points = np.array([n.p for n in part.fem.nodes])

    nodes_group = time_step.create_group("NOE")
    nodes_group.attrs.create("CGT", 1)
    nodes_group.attrs.create("CGS", 1)

    nodes_group.attrs.create("PFL", np.string_(profile))
    coo = nodes_group.create_dataset("COO", data=points.flatten(order="F"))
    coo.attrs.create("CGT", 1)
    coo.attrs.create("NBR", len(points))

    if _Settings.ca_experimental_id_numbering is True:
        node_ids = [n.id for n in part.fem.nodes]
        num = nodes_group.create_dataset("NUM", data=node_ids)
        num.attrs.create("CGT", 1)
        num.attrs.create("NBR", len(points))

    if len(part.fem.nsets.keys()) > 0:
        _add_node_sets(nodes_group, part, points, families)


def _write_elements(part, time_step, profile, families):
    """

    Add the following ['FAM', 'NOD', 'NUM'] to the 'MAI' group

    **NOD** requires 'CGT' and 'NBR' attrs

    :param part:
    :param time_step:
    :param profile:
    :param families:
    :return:
    """

    def get_node_ids_from_element(el_):
        return [int(n.id) for n in el_.nodes]

    elements_group = time_step.create_group("MAI")
    elements_group.attrs.create("CGT", 1)
    for group, elements in groupby(part.fem.elements, key=attrgetter("type")):
        if group in ElemShapes.masses + ElemShapes.springs:
            logging.error("NotImplemented: Skipping Mass or Spring Elements")
            continue
        med_type = abaqus_to_med_type(group)
        elements = list(elements)
        cells = np.array(list(map(get_node_ids_from_element, elements)))

        med_cells = elements_group.create_group(med_type)
        med_cells.attrs.create("CGT", 1)
        med_cells.attrs.create("CGS", 1)
        med_cells.attrs.create("PFL", np.string_(profile))

        nod = med_cells.create_dataset("NOD", data=cells.flatten(order="F"))
        nod.attrs.create("CGT", 1)
        nod.attrs.create("NBR", len(cells))

        # Node Numbering is necessary for proper handling of
        num = med_cells.create_dataset("NUM", data=[int(el.id) for el in elements])
        num.attrs.create("CGT", 1)
        num.attrs.create("NBR", len(cells))

    # Add Element sets
    if len(part.fem.elsets.keys()) > 0:
        _add_cell_sets(elements_group, part, families)


def _write_mesh_presets(f, mesh_name):
    """

    :param f:
    :param mesh_name:
    :return: Time step 0
    """
    numpy_void_str = np.string_("")
    dim = 3

    # Meshes
    mesh_ensemble = f.create_group("ENS_MAA")

    med_mesh = mesh_ensemble.create_group(mesh_name)
    med_mesh.attrs.create("DIM", dim)  # mesh dimension
    med_mesh.attrs.create("ESP", dim)  # spatial dimension
    med_mesh.attrs.create("REP", 0)  # cartesian coordinate system (repère in French)
    med_mesh.attrs.create("UNT", numpy_void_str)  # time unit
    med_mesh.attrs.create("UNI", numpy_void_str)  # spatial unit
    med_mesh.attrs.create("SRT", 1)  # sorting type MED_SORT_ITDT

    # component names:
    names = ["X", "Y", "Z"][:dim]
    med_mesh.attrs.create("NOM", np.string_("".join(f"{name:<16}" for name in names)))
    med_mesh.attrs.create("DES", np.string_("Mesh created with meshio"))
    med_mesh.attrs.create("TYP", 0)  # mesh type (MED_NON_STRUCTURE)

    # Time-step
    step = "-0000000000000000001-0000000000000000001"  # NDT NOR
    time_step = med_mesh.create_group(step)
    time_step.attrs.create("CGT", 1)
    time_step.attrs.create("NDT", -1)  # no time step (-1)
    time_step.attrs.create("NOR", -1)  # no iteration step (-1)
    time_step.attrs.create("PDT", -1.0)  # current time
    return time_step


def resolve_ids_in_multiple(tags, tags_data, is_elem):
    """
    Find elements shared by multiple sets

    :param tags:
    :param tags_data:
    :return:
    """
    from ada.fem import FemSet

    fin_data = dict()
    for t, memb in tags_data.items():
        fin_data[t] = []
        for mem in memb:
            refs = list(filter(lambda x: type(x) == FemSet, mem.refs))
            if len(refs) > 1:
                names = [r.name for r in refs]
                if names not in tags.values():
                    new_int = min(tags.keys()) - 1 if is_elem else max(tags.keys()) + 1
                    tags[new_int] = names
                    fin_data[new_int] = []
                else:
                    rmap = {tuple(v): r for r, v in tags.items()}
                    new_int = rmap[tuple(names)]
                if mem not in fin_data[new_int]:
                    fin_data[new_int].append(mem)
            else:
                fin_data[t].append(mem)
    to_be_removed = []
    for i, f in fin_data.items():
        if len(f) == 0:
            to_be_removed.append(i)
    for t in to_be_removed:
        fin_data.pop(t)
        tags.pop(t)
    return fin_data


def _add_cell_sets(cells_group, part, families):
    """

    :param cells_group:
    :param part:
    :param families:
    :type part: ada.Part
    """
    cell_id_num = -4

    element = families.create_group("ELEME")
    tags = dict()
    tags_data = dict()

    cell_id_current = cell_id_num
    for cell_set in part.fem.elsets.values():
        tags[cell_id_current] = [cell_set.name]
        tags_data[cell_id_current] = cell_set.members
        cell_id_current -= 1

    res_data = resolve_ids_in_multiple(tags, tags_data, True)

    def get_node_ids_from_element(el_):
        return [int(n.id - 1) for n in el_.nodes]

    for group, elements in groupby(part.fem.elements, key=attrgetter("type")):
        if group in ElemShapes.masses + ElemShapes.springs:
            logging.error("NotImplemented: Skipping Mass or Spring Elements")
            continue
        elements = list(elements)
        cell_ids = {el.id: i for i, el in enumerate(elements)}

        cell_data = np.zeros(len(elements), dtype=np.int32)

        for t, mem in res_data.items():
            list_filtered = [cell_ids[el.id] for el in filter(lambda x: x.type == group, mem)]
            for index in list_filtered:
                cell_data[index] = t

        cells = np.array(list(map(get_node_ids_from_element, elements)))
        med_type = abaqus_to_med_type(group)
        med_cells = cells_group.get(med_type)
        family = med_cells.create_dataset("FAM", data=cell_data)
        family.attrs.create("CGT", 1)
        family.attrs.create("NBR", len(cells))

    _write_families(element, tags)


def _add_node_sets(nodes_group, part, points, families):
    """
    :param nodes_group:
    :param part:
    :param families:
    :type part: ada.Part
    """
    tags = dict()

    # TODO: Simplify this function
    # tags_data = dict()
    # cell_id_current = 4
    # for cell_set in part.fem.nsets.values():
    #     tags[cell_id_current] = [cell_set.name]
    #     tags_data[cell_id_current] = cell_set.members
    #     cell_id_current += 1
    #
    # nmap = {n.id: i for i, n in enumerate(part.fem.nodes)}
    #
    # res_data = resolve_ids_in_multiple(tags, tags_data, False)
    # points = np.zeros(len(points), dtype=np.int32)
    #
    # for t, rd in res_data.items():
    #     for r in rd:
    #         points[nmap[r.id]] = t
    nsets = dict()
    for key, val in part.fem.nsets.items():
        nsets[key] = [int(p.id) for p in val]

    points = _set_to_tags(nsets, points, 2, tags)

    family = nodes_group.create_dataset("FAM", data=points)
    family.attrs.create("CGT", 1)
    family.attrs.create("NBR", len(points))

    # For point tags
    node = families.create_group("NOEUD")
    _write_families(node, tags)


def _resolve_element_in_use_by_other_set(tagged_data, ind, tags, name, is_elem):
    """


    :param tagged_data:
    :param ind:
    :param tags:
    :param name:
    :param is_elem:
    """
    existing_id = int(tagged_data[ind])
    current_tags = tags[existing_id]
    all_tags = current_tags + [name]

    if name in current_tags:
        raise ValueError("Unexpected error. Name already exists in set during resolving set members.")

    new_int = None
    for i_, t_ in tags.items():
        if all_tags == t_:
            new_int = i_
            break

    if new_int is None:
        new_int = int(min(tags.keys()) - 1) if is_elem else int(max(tags.keys()) + 1)
        tags[new_int] = tags[existing_id] + [name]

    tagged_data[ind] = new_int


def _set_to_tags(sets, data, tag_start_int, tags, id_map=None):
    """

    :param sets:
    :param data:
    :param tag_start_int:
    :param
    :return: The tagged data.
    """
    tagged_data = np.zeros(len(data), dtype=np.int32)
    tag_int = 0 + tag_start_int

    is_elem = False if tag_int > 0 else True

    tag_int = tag_start_int
    tag_map = dict()
    # Generate basic tags upfront
    for name in sets.keys():
        tags[tag_int] = [name]
        tag_map[name] = tag_int
        if is_elem is True:
            tag_int -= 1
        else:
            tag_int += 1

    for name, set_data in sets.items():
        if len(set_data) == 0:
            continue

        for index_ in set_data:
            index = int(index_ - 1)

            if id_map is not None:
                index = id_map[index_]

            if index > len(tagged_data) - 1:
                raise IndexError()

            if tagged_data[index] != 0:  # id is already defined in another set
                _resolve_element_in_use_by_other_set(tagged_data, index, tags, name, is_elem)
            else:
                tagged_data[index] = tag_map[name]

    return tagged_data


def _family_name(set_id, name):
    """Return the FAM object name corresponding to the unique set id and a list of
    subset names
    """
    return "FAM" + "_" + str(set_id) + "_" + "_".join(name)


def _write_families(fm_group, tags):
    """Write point/cell tag information under FAS/[mesh_name]"""
    for set_id, name in tags.items():
        family = fm_group.create_group(_family_name(set_id, name))
        family.attrs.create("NUM", set_id)
        group = family.create_group("GRO")
        group.attrs.create("NBR", len(name))  # number of subsets
        dataset = group.create_dataset("NOM", (len(name),), dtype="80int8")
        for i in range(len(name)):
            # make name 80 characters
            name_80 = name[i] + "\x00" * (80 - len(name[i]))
            # Needs numpy array, see <https://github.com/h5py/h5py/issues/1735>
            dataset[i] = np.array([ord(x) for x in name_80])



import numpy as np

from ada import Assembly, Beam, Material, Part, PrimBox, PrimCyl, PrimExtrude
from ada.fem import Bc, FemSet, Load, Step
from ada.fem.io.mesh.recipes import create_beam_mesh
from ada.fem.utils import get_beam_end_nodes


def beam_ex1(p1=(0, 0, 0), p2=(1.5, 0, 0), profile="IPE400"):
    """

    :return:
    :rtype: ada.Assembly
    """
    bm = Beam("MyBeam", p1, p2, profile, Material("S355"))
    a = Assembly("Test", creator="Kristoffer H. Andersen") / [Part("MyPart") / bm]

    h = 0.2
    r = 0.02

    normal = [0, 1, 0]
    xdir = [-1, 0, 0]

    # Polygon Extrusions
    origin = np.array([0.2, -0.1, -0.1])
    points = [(0, 0), (0.1, 0), (0.05, 0.1)]

    bm.add_penetration(PrimExtrude("Poly1", points, h, normal, origin, xdir))

    origin += np.array([0.2, 0, 0])
    points = [(0, 0, r), (0.1, 0, r), (0.05, 0.1, r)]

    bm.add_penetration(PrimExtrude("Poly2", points, h, normal, origin, xdir))

    origin += np.array([0.2, 0, 0])
    points = [(0, 0, r), (0.1, 0, r), (0.1, 0.2, r), (0.0, 0.2, r)]

    bm.add_penetration(PrimExtrude("Poly3", points, h, normal, origin, xdir))

    # Cylinder Extrude
    x = origin[0] + 0.2

    bm.add_penetration(PrimCyl("cylinder", (x, -0.1, 0), (x, 0.1, 0), 0.1))

    # Box Extrude
    x += 0.2

    bm.add_penetration(PrimBox("box", (x, -0.1, -0.1), (x + 0.2, 0.1, 0.1)))

    # Create a FEM analysis of the beam as a cantilever subjected to gravity loads
    p = a.get_part("MyPart")
    create_beam_mesh(bm, p.fem, "shell")

    # Add a set containing ALL elements (necessary for Calculix loads).
    fs = p.fem.add_set(FemSet("Eall", [el for el in p.fem.elements], "elset"))

    step = a.fem.add_step(Step("gravity", "static", nl_geom=True, init_incr=100.0, total_time=100.0))
    step.add_load(Load("grav", "gravity", -9.81 * 800, fem_set=fs))

    fix_set = p.fem.add_set(FemSet("bc_nodes", get_beam_end_nodes(bm), "nset"))
    a.fem.add_bc(Bc("Fixed", fix_set, [1, 2, 3]))
    return a



import logging
import mmap
import os
import pathlib
import re
from itertools import chain

import numpy as np

from ada.core.containers import Nodes
from ada.core.utils import Counter, roundoff
from ada.fem import (
    FEM,
    Bc,
    Connector,
    ConnectorSection,
    Constraint,
    Csys,
    Elem,
    ElemShapes,
    FemSection,
    FemSet,
    Interaction,
    InteractionProperty,
    Mass,
    PredefinedField,
    Surface,
)
from ada.fem.containers import FemElements, FemSections, FemSets
from ada.fem.io.abaqus.common import AbaCards
from ada.materials.metals import CarbonSteel

from ..utils import str_to_int

part_name_counter = Counter(1, "Part")
_re_in = re.IGNORECASE | re.MULTILINE | re.DOTALL


def read_fem(assembly, fem_file, fem_name=None):
    """
    This will create and add an AbaqusPart object based on a path reference to a Abaqus input file.

    :param assembly: Assembly object
    :param fem_file: Absolute or relative path to Abaqus *.inp file. See below for further explanation
    :param fem_name: Currently not in use.
    :type assembly: ada.Assembly
    """
    print("Starting import of Abaqus input file")

    if fem_name is not None:
        global part_name_counter
        part_name_counter = Counter(1, fem_name)

    bulk_str = read_bulk_w_includes(fem_file)
    lbulk = bulk_str.lower()
    ass_start = lbulk.find("\n*assembly")
    ass_end = lbulk.rfind("\n*end assembly")
    step_start = lbulk.rfind("\n*step")

    ass_start = ass_start + 1 if ass_start != -1 else ass_start
    ass_end = ass_end + 1 if ass_end != -1 else ass_end
    step_start = step_start + 1 if step_start != -1 else step_start

    if ass_start == -1 and ass_end == -1:
        uses_assembly_parts = False
        assembly_str = bulk_str
        props_str = bulk_str
    else:
        uses_assembly_parts = True
        assembly_str = bulk_str[ass_start : ass_end + 2]
        props_str = bulk_str[ass_end + 2 : step_start]

    inst_end = assembly_str.lower().rfind("\n*end instance")
    inst_end = inst_end + 15 if inst_end != -1 else inst_end
    get_materials_from_bulk(assembly, props_str)
    get_intprop_from_lines(assembly, props_str)

    inst_matches = re.compile(r"\*Instance, name=(.*?), part=(.*?)\n(.*?)\*End Instance", _re_in)
    ass_data = {}
    for ad in map(get_instance_data, inst_matches.finditer(assembly_str[:inst_end])):
        if ad[0] not in ass_data.keys():
            ass_data[ad[0]] = []
        ass_data[ad[0]].append(ad[1:])

    import_parts(assembly, bulk_str[:ass_start], ass_data)

    ass_sets = assembly_str[inst_end:]

    ass_fem = assembly.fem

    if uses_assembly_parts is True:
        assembly.fem._nodes += get_nodes_from_inp(ass_sets, ass_fem)
        assembly.fem.lcsys.update(get_lcsys_from_bulk(ass_sets, ass_fem))
        assembly.fem._elements += get_elem_from_inp(ass_sets, ass_fem)
        assembly.fem._sets += get_sets_from_bulk(ass_sets, ass_fem)
        assembly.fem.sets.link_data()
        assembly.fem.surfaces.update(get_surfaces_from_bulk(ass_sets, ass_fem))
        assembly.fem._constraints += get_constraints_from_inp(ass_sets, ass_fem)
        assembly.fem.connector_sections.update(get_connector_sections_from_bulk(props_str, ass_fem))
        assembly.fem.connectors.update(get_connectors_from_inp(ass_sets, ass_fem))
        assembly.fem._bcs += get_bcs_from_bulk(props_str, ass_fem)

    get_interactions_from_bulk_str(props_str, assembly)
    get_initial_conditions_from_lines(assembly, props_str)


def read_bulk_w_includes(inp_path):
    """

    :param inp_path: Absolute path to input file
    :return: bulk str
    """

    re_bulk_include = re.compile(r"\*Include,\s*input=(.*?)$", _re_in)
    bulk_repl = dict()
    with open(inp_path, "r") as inpDeck:
        bulk_str = inpDeck.read()
        for m in re_bulk_include.finditer(bulk_str):
            search_key = m.group(0)
            with open(os.path.join(os.path.dirname(inp_path), m.group(1)), "r") as d:
                bulk_repl[search_key] = d.read()

    for key, val in bulk_repl.items():
        bulk_str = bulk_str.replace(key, val)
    return bulk_str


def import_bulk(file_path, buffer_function):
    with open(file_path, "r") as f:
        return buffer_function(f.read())


def import_bulk2(file_path, buffer_function):
    with open(file_path, "r") as f:
        with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as m:
            return buffer_function(m)


def import_parts(assembly, bulk_str, ass_data):
    """
    Import parts from an assembly

    :param assembly:
    :param bulk_str:
    :param ass_data:
    :type assembly: ada.Assembly
    """

    parts_matches = re.compile(r"\*Part, name=(.*?)\n(.*?)\*End Part", _re_in)
    part_names = re.compile(r"\*\*\s*PART INSTANCE:\s*(.*?)\n(.*)", _re_in)

    def grab_instance_data(name):
        if name in ass_data:
            return ass_data[name]
        else:
            return None, None, None

    def get_part(m):
        name = m.group(1)
        part_bulk_str = m.group(2)
        instances = grab_instance_data(name)
        parts = []
        for instance in instances:
            inst_name, inst_bulk, inst_metadata = instance
            part_bulk_str = inst_bulk if part_bulk_str == "" and inst_bulk != "" else part_bulk_str
            part = get_part_from_bulk_str(name, part_bulk_str, assembly, inst_name, inst_metadata)
            parts.append(part)
        return parts

    def get_part_without_assembly():
        part_name_matches = list(part_names.finditer(bulk_str))
        p_nmatch = tuple(part_name_matches)

        if len(p_nmatch) != 1:
            p_bulk = bulk_str
            p_name = None
        else:
            p_name = p_nmatch[0].group(1)
            p_bulk = p_nmatch[0].group(2)

        p_name = next(part_name_counter) if p_name is None else p_name
        return get_part_from_bulk_str(p_name, p_bulk, assembly)

    # import concurrent.futures
    # with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
    #     futures = []
    #     for number in parts_matches.finditer(bulk_str):
    #         future = executor.submit(get_part, number)
    #         futures.append(future)
    #
    #     if len(futures) == 0:
    #         future = executor.submit(get_part_without_assembly)
    #         futures.append(future)
    #
    #     for future in concurrent.futures.as_completed(futures):
    #         p = future.result()
    #         self.assembly.add_part(p)
    #         print(8 * '-' + f'Imported "{p.fem.instance_name}"')
    #
    new_parts = 0
    from itertools import count

    ncounter = count(1)
    part_list = list(chain.from_iterable(map(get_part, parts_matches.finditer(bulk_str))))
    for p in part_list:
        if p.name is None:
            p.name = f"Part{next(ncounter)}"
        if p.name in assembly.parts.keys():
            p.name = p.fem.name
        assembly.add_part(p)
        new_parts += 1
        print(8 * "-" + f'Imported "{p.fem.instance_name}"')
    if new_parts == 0:
        p = get_part_without_assembly()
        if p.name is None:
            p.name = f"Part{next(ncounter)}"
        assembly.add_part(p)
        print(8 * "-" + f'Imported "{p.fem.instance_name}"')


def get_part_from_bulk_str(name, bulk_str, parent, instance_name=None, metadata=None):
    from ada import Part

    metadata = dict(move=None, rotate=None) if metadata is None else metadata
    instance_name = name if instance_name is None else instance_name
    if instance_name is None:
        instance_name = "Temp"
    fem = FEM(name=instance_name, metadata=metadata)
    part = Part(name, fem=fem, parent=parent)
    fem._nodes = get_nodes_from_inp(bulk_str, fem)
    fem.nodes.move(move=fem.metadata["move"], rotate=fem.metadata["rotate"])
    fem._elements = get_elem_from_inp(bulk_str, fem)
    fem.elements.build_sets()
    fem._sets = fem.sets + get_sets_from_bulk(bulk_str, fem)
    fem._sections = get_sections_from_inp(bulk_str, fem)
    fem._bcs += get_bcs_from_bulk(bulk_str, fem)
    fem._masses = get_mass_from_bulk(bulk_str, fem)
    fem.surfaces.update(get_surfaces_from_bulk(bulk_str, fem))
    fem._lcsys = get_lcsys_from_bulk(bulk_str, fem)
    fem._constraints = get_constraints_from_inp(bulk_str, fem)

    return part


def get_initial_conditions_from_lines(assembly, bulk_str):
    """
    TODO: Optimize this function


    ** PREDEFINED FIELDS
    **
    ** Name: IC-1   Type: Velocity
    *Initial Conditions, type=VELOCITY
    CONTAINER20FT-10000KG, 1, 0.
    CONTAINER20FT-10000KG, 2, 0.
    CONTAINER20FT-10000KG, 3, 4.1
    """
    if bulk_str.find("*Initial Conditions") == -1:
        return
    re_str = r"(?:\*\*\s*Name:\s*(.*?)\s*Type:\s*(.*?)\n)+\*Initial Conditions, type=.*?\n(?<=)((?:.*?)(?=\*|\Z))"

    def sort_props(line):
        ev = [x.strip() for x in line.split(",")]
        set_name = ev[0]
        try:
            dofs = int(ev[1])
        except BaseException as e:
            logging.debug(e)
            dofs = ev[1]
        if len(ev) > 2:
            magn = ev[2]
        else:
            magn = None
        return set_name, dofs, magn

    def grab_init_props(m):
        bc_name = m.group(1)
        bc_type = m.group(2)
        bcs = m.group(3)
        props = [sort_props(line) for line in bcs.splitlines()]
        set_name, dofs, magn = list(zip(*props))
        fem_set = None
        set_name_up = set_name[0]
        if set_name_up in assembly.fem.elsets.keys():
            fem_set = assembly.fem.elsets[set_name_up]
        elif set_name_up in assembly.fem.nsets.keys():
            fem_set = assembly.fem.nsets[set_name_up]
        else:
            for p in assembly.get_all_parts_in_assembly():
                if set_name_up in p.fem.elsets.keys():
                    fem_set = p.fem.elsets[set_name_up]
        if fem_set is None:
            raise ValueError(f'Unable to find fem set "{set_name[0]}"')

        return PredefinedField(bc_name, bc_type, fem_set, dofs, magn, parent=assembly.fem)

    for match in re.finditer(re_str, bulk_str, _re_in):
        assembly.fem.add_predefined_field(grab_init_props(match))


def get_materials_from_bulk(assembly, bulk_str):
    re_str = (
        r"(\*Material,\s*name=.*?)(?=\*|\Z)(?!\*Elastic|\*Density|\*Plastic|"
        r"\*Damage Initiation|\*Damage Evolution|\*Expansion)"
    )
    re_materials = re.compile(re_str, _re_in)
    for m in re_materials.finditer(bulk_str):
        mat = mat_str_to_mat_obj(m.group())
        assembly.add_material(mat)


def get_intprop_from_lines(assembly, bulk_str):
    """
    *Surface Interaction, name=contactProp
    *Friction
    0.,
    *Surface Behavior, pressure-overclosure=HARD
    """
    re_str = r"(\*Surface Interaction, name=.*?)(?=\*)(?!\*Friction|\*Surface Behavior|\*Surface Smoothing)"
    # surf_interact = AbaFF("Surface Interaction", [("name=",), ("bulk>",)], [("Friction", [(), ("bulkFRIC>",)])])

    assembly.fem.metadata["surf_smoothing"] = []
    for m in AbaCards.surface_smoothing.regex.finditer(bulk_str):
        d = m.groupdict()
        assembly.fem.metadata["surf_smoothing"].append(d)

    for m in re.findall(re_str, bulk_str, _re_in):
        name = re.search(r"name=(.*?)\n", m, _re_in).group(1)
        re_fric = re.search(r"\*Friction\n(.*?),", m, _re_in)
        fric = None
        if re_fric is not None:
            fric = re_fric.group(1)
        props = dict(name=name, friction=fric)
        res = re.search(
            r"\*Surface Behavior,\s*(?P<btype>.*?)(?:=(?P<behaviour>.*?)(?:$\n(?P<tabular>.*?)(?:\*|\Z)|$)|$)",
            m,
            _re_in,
        )
        if res is not None:
            res = res.groupdict()
            btype = res["btype"]
            behave = res["behaviour"] if btype.upper() != "PENALTY" else btype
            tabular = res["tabular"]
            if tabular is not None:
                tab_list = []
                for line in tabular.splitlines():
                    tab_list.append(tuple(np.fromstring(line, dtype=float, sep=",")))
                tabular = tab_list
                props.update(dict(pressure_overclosure=behave, tabular=tabular))
        assembly.fem.add_interaction_property(InteractionProperty(**props))


def get_instance_data(m):
    """

    Move/rotate data lines are specified here:

    https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-instance.htm

    """

    move_rot = re.compile(r"(?:^\s*(.*?),\s*(.*?),\s*(.*?)$)", _re_in)
    metadata = dict(move=None, rotate=None)
    inst_name = m.group(1)
    p_ref = m.group(2)
    inst_bulk = m.group(3)
    move = None
    rotate = None
    mr = move_rot.finditer(inst_bulk)
    if mr is not None:
        for j, mo in enumerate(mr):
            content = mo.group(0)
            if "*" in content or j == 2 or content == "":
                break
            if j == 0:
                move = [float(mo.group(1)), float(mo.group(2)), float(mo.group(3))]
            if j == 1:
                r = [float(x) for x in mo.group(3).split(",")]
                rotate = [
                    float(mo.group(1)),
                    float(mo.group(2)),
                    r[0],
                    r[1],
                    r[2],
                    r[3],
                    r[4],
                ]
            if move is not None:
                metadata["move"] = tuple(move)
            if rotate is not None:
                metadata["rotate"] = (
                    tuple(rotate[:3]),
                    tuple(rotate[3:-1]),
                    rotate[-1],
                )
    return p_ref, inst_name, inst_bulk, metadata


def mat_str_to_mat_obj(mat_str):
    """
    Converts a Abaqus materials str into a ADA Materials object

    :param mat_str:
    :return:
    """
    from ada import Material

    rd = roundoff

    # Name
    name = re.search(r"name=(.*?)\n", mat_str, _re_in).group(1).split("=")[-1].strip()

    # Density
    density_ = re.search(r"\*Density\n(.*?)(?:,|$)", mat_str, _re_in)
    if density_ is not None:
        density = rd(density_.group(1).strip().split(",")[0].strip(), 10)
    else:
        print('No density flag found for material "{}"'.format(name))
        density = None

    # Elastic
    re_elastic_ = re.search(r"\*Elastic(?:,\s*type=(.*?)|)\n(.*?)(?:\*|$)", mat_str, _re_in)
    if re_elastic_ is not None:
        re_elastic = re_elastic_.group(2).strip().split(",")
        young, poisson = rd(re_elastic[0]), rd(re_elastic[1])
    else:
        print('No Elastic properties found for material "{name}"'.format(name=name))
        young, poisson = None, None

    # Plastic
    re_plastic_ = re.search(r"\*Plastic\n(.*?)(?:\*|\Z)", mat_str, _re_in)
    if re_plastic_ is not None:
        re_plastic = [tuple(x.split(",")) for x in re_plastic_.group(1).strip().splitlines()]
        sig_p = [rd(x[0]) for x in re_plastic]
        eps_p = [rd(x[1]) for x in re_plastic]
    else:
        eps_p, sig_p = None, None

    # Expansion
    re_zeta = re.search(r"\*Expansion(?:,\s*type=(.*?)|)\n(.*?)(?:\*|$)", mat_str, _re_in)
    if re_zeta is not None:
        zeta = float(re_zeta.group(2).split(",")[0].strip())
    else:
        zeta = 0.0

    # Return material object
    model = CarbonSteel(
        rho=density,
        E=young,
        v=poisson,
        eps_p=eps_p,
        zeta=zeta,
        sig_p=sig_p,
        plasticity_model=None,
    )
    return Material(name=name, mat_model=model)


def import_multiple_inps(input_files_dir):
    """
    Import a set of inp files from a folder

    :param input_files_dir:
    """

    def read_inp(fname):
        if ".inp" not in fname:
            return None
        else:
            with open(pathlib.Path(input_files_dir) / fname, "r") as d:
                return d.read()

    return "".join([x for x in map(read_inp, os.listdir(input_files_dir)) if x is not None])


def get_nodes_from_inp(bulk_str, parent):
    """
    Extract node information from abaqus input file string

    """
    from ada import Node

    re_no = re.compile(
        r"^\*Node\s*(?:,\s*nset=(?P<nset>.*?)\n|\n)(?P<members>(?:.*?)(?=\*|\Z))",
        _re_in,
    )

    def getnodes(m):
        d = m.groupdict()
        res = np.fromstring(list_cleanup(d["members"]), sep=",", dtype=np.float64)
        res_ = res.reshape(int(res.size / 4), 4)
        members = [Node(n[1:4], int(n[0]), parent=parent) for n in res_]
        if d["nset"] is not None:
            parent.sets.add(FemSet(d["nset"], members, "nset", parent=parent))
        return members

    nodes = list(chain.from_iterable(map(getnodes, re_no.finditer(bulk_str))))

    return Nodes(
        nodes,
        parent=parent,
    )


def get_elem_from_inp(bulk_str, fem):
    """
    Extract elements from abaqus input file

    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM

    :return:
    :rtype: ada.fem.containers.FemElements
    """

    from ada import Node, Part

    re_el = re.compile(
        r"^\*Element,\s*type=(?P<eltype>.*?)(?:\n|,\s*elset=(?P<elset>.*?)\s*\n)(?<=)(?P<members>(?:.*?)(?=\*|\Z))",
        _re_in,
    )

    def grab_elements(match):
        d = match.groupdict()
        eltype = d["eltype"]
        elset = d["elset"]
        members = d["members"]
        res = re.search("[a-zA-Z]", members)
        if eltype.upper() in ElemShapes.cube20 + ElemShapes.cube27 or res is None:
            if eltype.upper() in ElemShapes.cube20 + ElemShapes.cube27:
                temp = members.splitlines()
                ntext = "".join([l1.strip() + "    " + l2.strip() + "\n" for l1, l2 in zip(temp[:-1:2], temp[1::2])])
            else:
                ntext = d["members"]
            res = np.fromstring(ntext.replace("\n", ","), sep=",", dtype=int)
            n = ElemShapes.num_nodes(eltype) + 1
            res_ = res.reshape(int(res.size / n), n)
            return [Elem(e[0], [fem.nodes.from_id(n) for n in e[1:]], eltype, elset, parent=fem) for e in res_]
        else:

            # TODO: This code needs to be re-worked!
            elems = []
            for li in members.splitlines():
                new_mem = []
                temp = li.split(",")
                elid = str_to_int(temp[0])
                for d in temp[1:]:
                    temp2 = [x.strip() for x in d.split(".")]
                    par_ = None
                    if len(temp2) == 2:
                        par, setr = temp2
                        pfems = []
                        parents = fem.parent.get_all_parts_in_assembly()
                        for p in parents:
                            assert isinstance(p, Part)
                            pfems.append(p.fem.name)
                            if p.fem.name == par:
                                par_ = p
                                break
                        if par_ is None:
                            raise ValueError(f'Unable to find parent for "{par}"')
                        r = par_.fem.nodes.from_id(str_to_int(setr))
                        if type(r) != Node:
                            raise ValueError("Node ID not found")
                        new_mem.append(r)
                    else:
                        r = fem.nodes.from_id(str_to_int(d))
                        if type(r) != Node:
                            raise ValueError("Node ID not found")
                        new_mem.append(r)
                elems.append(Elem(elid, new_mem, eltype, elset, parent=fem))
            return elems

    return FemElements(
        chain.from_iterable(map(grab_elements, re_el.finditer(bulk_str))),
        fem_obj=fem,
    )


def get_sections_from_inp(bulk_str, fem):
    """

    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM
    :return:
    """
    iter_beams = get_beam_sections_from_inp(bulk_str, fem)
    iter_shell = get_shell_sections_from_inp(bulk_str, fem)
    iter_solid = get_solid_sections_from_inp(bulk_str, fem)

    return FemSections(chain.from_iterable([iter_beams, iter_shell, iter_solid]), fem)


def get_sets_from_bulk(bulk_str, fem):
    """

    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM
    :return: Collection of sets
    :rtype: FemSets
    """
    from ada import Assembly

    re_sets = re.compile(
        r"(?:\*(nset|elset),\s*(?:nset|elset))=(.*?)(?:,\s*(internal\s*)|(?:))"
        r"(?:,\s*instance=(.*?)|(?:))(?:,\s*(generate)|(?:))\s*\n(?<=)((?:.*?)(?=\*|\Z))",
        _re_in,
    )

    all_parts = fem.parent.get_all_parts_in_assembly()

    def str_to_ints(instr):
        try:
            return [int(x.strip()) for l in instr.splitlines() for x in l.split(",") if x.strip() != ""]
        except ValueError:
            return [str(x.strip()) for l in instr.splitlines() for x in l.split(",") if x.strip() != ""]

    def get_parent_instance(instance):
        """

        :rtype: ada.fem.FEM
        """
        if instance is None:
            return fem
        elif instance is not None and type(fem.parent) is Assembly:
            for p in filter(lambda x: x.fem.instance_name == instance, all_parts):
                return p.fem
        else:
            raise ValueError(f'Unable to find instance "{instance}" amongst assembly parts')

    def get_set(match):
        name = match.group(2)
        set_type = match.group(1)
        internal = True if match.group(3) is not None else False
        instance = match.group(4)
        generate = True if match.group(5) is not None else False
        members_str = match.group(6)
        gen_mem = str_to_ints(members_str) if generate is True else []
        members = [] if generate is True else str_to_ints(members_str)
        metadata = dict(instance=instance, internal=internal, generate=generate, gen_mem=gen_mem)

        parent_instance = get_parent_instance(instance)

        fem_set = FemSet(
            name,
            members,
            set_type=set_type,
            metadata=metadata,
            parent=parent_instance,
        )
        return fem_set

    return FemSets(list(map(get_set, re_sets.finditer(bulk_str))), fem)

    # import concurrent.futures
    # with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
    #     futures = []
    #     for number in re_sets.finditer(bulk_str):
    #         future = executor.submit(get_set, number)
    #         futures.append(future)
    #     sets = [future.result() for future in concurrent.futures.as_completed(futures)]
    #
    # return FemSetsCollection(sets, parent)


def get_bcs_from_bulk(bulk_str, fem):
    """

    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM
    :return:
    """

    bc_counter = Counter(1, "bc")

    re_bcs = re.compile(
        r"(?:\*\*\s*Name:\s*(?P<name>.*?)\s*Type:\s*(?P<type>.*?)\n|)"
        r"\*Boundary\n(?<=)(?P<content>(?:.*?)(?=\*|\Z))",
        _re_in,
    )

    def get_dofs(content):
        set_name = None
        dofs = []
        magn = []
        if content.count("\n") == 1:
            temp = content.split(",")
            set_name = temp[0].strip()
            dof_in = temp[1].strip().replace("\n", "")
            if dof_in.lower() == "encastre":
                dofs = dof_in
            else:
                dof = str_to_int(temp[1])
                if len(temp) == 2:
                    dofs = [x if dof == x else None for x in range(1, 7)]
                else:
                    dof_end = str_to_int(temp[2])
                    dofs = [x if dof <= x <= dof_end else None for x in range(1, 7)]

        else:
            for line in content.splitlines():
                ev = [x.strip() for x in line.split(",")]
                set_name = ev[0]
                try:
                    dofs.append(int(ev[1]))
                except BaseException as e:
                    logging.debug(e)
                    dofs.append(ev[1])
                if len(ev) > 3:
                    magn.append(ev[2])
        magn = None if len(magn) == 0 else magn
        return set_name, dofs, magn

    def get_nset(part_instance_name, set_name):
        for p in fem.parent.parts.values():
            if p.fem.instance_name == part_instance_name:
                return p.fem.sets.get_nset_from_name(set_name)

    def get_bc(match):
        d = match.groupdict()
        bc_name = d["name"] if d["name"] is not None else next(bc_counter)
        bc_type = d["type"]
        set_name, dofs, magn = get_dofs(d["content"])

        if "." in set_name:
            part_instance_name, set_name = set_name.split(".")
            fem_set = get_nset(part_instance_name, set_name)
        else:
            if set_name in fem.nsets.keys():
                fem_set = fem.sets.get_nset_from_name(set_name)
            else:
                val = str_to_int(set_name)
                if val in fem.nodes.dmap.keys():
                    node = fem.nodes.from_id(val)
                    fem_set = FemSet(bc_name + "_set", [node], "nset", parent=fem)
                else:
                    raise ValueError(f'Unable to find set "{set_name}" in part {fem}')

        if fem_set is None:
            raise Exception("Unable to Find node set")

        props = dict()
        if bc_type is not None:
            props["bc_type"] = bc_type
        if magn is not None:
            props["magnitudes"] = magn

        return Bc(bc_name, fem_set, dofs, parent=fem, **props)

    return [get_bc(match_in) for match_in in re_bcs.finditer(bulk_str)]


def get_mass_from_bulk(bulk_str, parent):
    """

    *MASS,ELSET=MASS3001
    2.00000000E+03,

    :return:
    """

    re_masses = re.compile(
        r"\*(?P<mass_type>Nonstructural Mass|Mass|Rotary Inertia),\s*elset=(?P<elset>.*?)"
        r"(?:,\s*type=(?P<ptype>.*?)\s*|\s*)(?:, units=(?P<units>.*?)|\s*)\n\s*(?P<mass>.*?)$",
        _re_in,
    )

    def map_element(el_set, mass_prop):
        """

        :param el_set:
        :param mass_prop:
        :type el_set: ada.fem.FemSet
        """
        elem = el_set.members[0]
        elem.mass_prop = mass_prop

    def get_mass(match):
        d = match.groupdict()
        elset = parent.sets.get_elset_from_name(d["elset"])
        mass_type = d["mass_type"]
        p_type = d["ptype"]
        mass = [str_to_int(x.strip()) for x in d["mass"].split(",") if x.strip() != ""]
        units = d["units"]
        mass = Mass(d["elset"], elset, mass, mass_type, p_type, units, parent=parent)
        map_element(elset, mass)
        return mass

    return {m.name: m for m in map(get_mass, re_masses.finditer(bulk_str))}


def get_surfaces_from_bulk(bulk_str, parent):
    """

    *Surface, type=ELEMENT, name=btn_surf
    _btn_surf_SPOS, SPOS

    :return:
    """

    def interpret_member(mem):
        msplit = mem.split(",")
        try:
            ref = str_to_int(msplit[0])
        except BaseException as e:
            logging.debug(e)
            ref = msplit[0].strip()

        return tuple([ref, msplit[1].strip()])

    surf_d = dict()
    for m in AbaCards.surface.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"].strip()
        surf_type = d["type"] if d["type"] is not None else "ELEMENT"
        members_str = d["bulk"]
        if members_str.count("\n") >= 1:
            id_refs = [interpret_member(m) for m in members_str.splitlines()]
            set_ref, set_id_ref = None, None
        else:
            id_refs = None
            res = [x.strip() for x in members_str.split(",")]
            if len(res) == 2:
                set_ref, set_id_ref = res
            else:
                set_ref = res[0]
                set_id_ref = 1.0

        if id_refs is None:
            if surf_type == "NODE":
                if set_id_ref == "":
                    fem_set = FemSet(f"n{set_ref}_set", [int(set_ref)], "nset")
                    parent.add_set(fem_set)
                    weight_factor = 1.0
                else:
                    if "." in set_ref:
                        ssplit = set_ref.split(".")
                        parent_ = None
                        fem_set = None
                        for prt in parent.parent.get_all_parts_in_assembly():
                            if prt.fem.name == ssplit[0]:
                                parent_ = prt.fem
                                fem_set = prt.fem.nsets[ssplit[1]]
                                break
                        if parent_ is None:
                            raise ValueError(f'Unable to find parent FEM "{ssplit[0]}"')
                    else:
                        fem_set = parent.nsets[set_ref]
                    weight_factor = float(set_id_ref)
                face_id_label = None
            else:
                weight_factor = None
                face_id_label = set_id_ref
                fem_set = parent.sets.get_elset_from_name(set_ref)
        else:
            fem_set = None
            weight_factor = None
            face_id_label = None
        surf_d[name] = Surface(
            name,
            surf_type,
            fem_set,
            weight_factor,
            face_id_label,
            id_refs,
            parent=parent,
        )

    return surf_d


def get_lcsys_from_bulk(bulk_str, parent):
    """
    https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-orientation.htm#simakey-r-orientation


    :param bulk_str:
    :param parent:
    :return:
    """
    # re_lcsys = re.compile(
    #     r"^\*Orientation(:?,\s*definition=(?P<definition>.*?)|)(?:,\s*system=(?P<system>.*?)|)\s*name=(?P<name>.*?)\n(?P<content>.*?)$",
    #     _re_in,
    # )

    lcsysd = dict()
    for m in AbaCards.orientation.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"].replace('"', "")
        defi = d["definition"] if d["definition"] is not None else "COORDINATES"
        system = d["system"] if d["system"] is not None else "RECTANGULAR"
        if defi.upper() == "COORDINATES":
            coords = [
                (float(d["ax"]), float(d["ay"]), float(d["az"])),
                (float(d["bx"]), float(d["by"]), float(d["bz"])),
            ]
            if d["cx"] is not None:
                coords += [(float(d["cx"]), float(d["cy"]), float(d["cz"]))]
            lcsysd[name] = Csys(name, system=system, coords=coords, parent=parent)
        else:
            raise NotImplementedError(f'Orientation definition "{defi}" is not yet supported')

    return lcsysd


def get_constraints_from_inp(bulk_str, fem):
    """

    ** Constraint: Container_RigidBody
    *Rigid Body, ref node=container_rp, elset=container

    *MPC
     BEAM,    2007,     161
     BEAM,    2008,     162

    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM
    """

    # Rigid Bodies

    constraints = []
    rbnames = Counter(1, "rgb")
    conames = Counter(1, "co")

    for m in AbaCards.tie.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"]
        msurf = grab_set_from_assembly(d["surf1"], fem, "surface")
        ssurf = grab_set_from_assembly(d["surf2"], fem, "surface")
        constraints.append(Constraint(name, "tie", msurf, ssurf, metadata=dict(adjust=d["adjust"])))

    for m in AbaCards.rigid_bodies.regex.finditer(bulk_str):
        d = m.groupdict()
        name = next(rbnames)
        ref_node = grab_set_from_assembly(d["ref_node"], fem, "nset")
        elset = grab_set_from_assembly(d["elset"], fem, "elset")
        constraints.append(Constraint(name, "rigid body", ref_node, elset, parent=fem))

    couplings = []
    for m in AbaCards.coupling.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["constraint_name"]
        rn = d["ref_node"].strip()
        sf = d["surface"].strip()
        if rn.isnumeric():
            ref_set = FemSet(next(conames), [int(rn)], "nset", parent=fem)
            fem.sets.add(ref_set)
        else:
            ref_set = fem.nsets[rn]

        surf = fem.surfaces[sf]

        res = np.fromstring(list_cleanup(d["bulk"]), sep=",", dtype=int)
        size = res.size
        cols = 2
        rows = int(size / cols)
        dofs = res.reshape(rows, cols)

        csys_name = d.get("orientation", None)
        if csys_name is not None:
            if csys_name not in fem.lcsys.keys():
                raise ValueError(f'Csys "{csys_name}" was not found on part {fem}')
            csys = fem.lcsys[csys_name]
        else:
            csys = None

        couplings.append(Constraint(name, "coupling", ref_set, surf, csys=csys, dofs=dofs, parent=fem))

    # Shell to Solid Couplings
    sh2solids = []
    for m in AbaCards.sh2so_re.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["constraint_name"]
        surf1 = grab_set_from_assembly(d["surf1"], fem, "surface")
        surf2 = grab_set_from_assembly(d["surf2"], fem, "surface")
        sh2solids.append(Constraint(name, "shell2solid", surf1, surf2))

    # MPC's
    mpc_dict = dict()
    mpc_re = re.compile(r"\*mpc\n(?P<bulk>.*?)^\*", _re_in)
    mpc_content = re.compile(r"\s*(?P<mpc_type>.*?),\s*(?P<master>.*?),\s*(?P<slave>.*?)$", _re_in)
    for match in mpc_re.finditer(bulk_str):
        d1 = match.groupdict()
        for subm in mpc_content.finditer(d1["bulk"]):
            d = subm.groupdict()
            mpc_type = d["mpc_type"]
            m = d["master"]
            s = d["slave"]
            if mpc_type not in mpc_dict.keys():
                mpc_dict[mpc_type] = []
            try:
                n1_ = str_to_int(m)
            except BaseException as e:
                logging.debug(e)
                n1_ = grab_set_from_assembly(m, fem, "nset")

            try:
                n2_ = str_to_int(s)
            except BaseException as e:
                logging.debug(e)
                n2_ = grab_set_from_assembly(s, fem, "nset")

            mpc_dict[mpc_type].append((n1_, n2_))

    def get_mpc(mpc_values):
        m_set, s_set = zip(*mpc_values)
        mpc_name = mpc_type + "_mpc"
        mset = FemSet("mpc_" + mpc_type + "_m", m_set, "nset")
        sset = FemSet("mpc_" + mpc_type + "_s", s_set, "nset")
        return Constraint(mpc_name, "mpc", mset, sset, mpc_type=mpc_type, parent=fem)

    mpcs = [get_mpc(mpc_values_in) for mpc_values_in in mpc_dict.values()]

    return list(chain.from_iterable([constraints, couplings, sh2solids, mpcs]))


def get_connector_sections_from_bulk(bulk_str, parent):
    """

    :param bulk_str:
    :param parent:
    :return:
    """

    consecsd = dict()

    for m in AbaCards.connector_behaviour.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"]
        comp = int(d["component"])

        res = np.fromstring(list_cleanup(d["bulk"]), sep=",", dtype=np.float64)
        size = res.size
        cols = comp + 1
        rows = int(size / cols)
        res_ = res.reshape(rows, cols)
        consecsd[name] = ConnectorSection(name, [res_], [], metadata=d)
    return consecsd


def get_connectors_from_inp(bulk_str, fem):
    """

    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM
    :return:
    """

    nsuffix = Counter(1, "_")
    cons = dict()
    for m in AbaCards.connector_section.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["behavior"] + next(nsuffix)
        elset = fem.elsets[d["elset"]]
        elem = elset.members[0]

        csys_ref = d["csys"].replace('"', "")
        csys_ref = csys_ref[:-1] if csys_ref[-1] == "," else csys_ref

        con_type = d["contype"]
        if con_type[-1] == ",":
            con_type = con_type[:-1]

        n1 = elem.nodes[0]
        n2 = elem.nodes[1]

        cons[name] = Connector(
            name,
            elem.id,
            n1,
            n2,
            con_type,
            fem.connector_sections[d["behavior"]],
            csys=fem.lcsys[csys_ref],
        )
    return cons


def get_beam_sections_from_inp(bulk_str, fem):
    """

    https://abaqus-docs.mit.edu/2017/English/SIMACAEELMRefMap/simaelm-c-beamcrosssectlib.htm

    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM
    """
    from ada import Section
    from ada.sections import GeneralProperties

    if bulk_str.lower().find("*beam section") == -1:
        return []
    re_beam = re.compile(
        r"(?:\*\*\s*Section:\s*(?P<sec_name>.*?)\s*Profile:\s*(?P<profile_name>.*?)\n|)"
        r"\*Beam Section,\s*elset=(?P<elset>.*?)\s*,\s*material=(?P<material>.*?)\s*,\s*"
        r"(?:temperature=(?P<temperature>.*?),|)\s*(?:section=|sect=)(?P<sec_type>.*?)\n"
        r"(?P<line1>.*?)\n(?P<line2>.*?)$",
        _re_in,
    )

    def interpret_section(profile_name, sec_type, props):
        props_clean = [roundoff(x) for x in filter(lambda x: x.strip() != "", props.split(","))]
        if sec_type.upper() == "BOX":
            b, h, t1, t2, t3, t4 = props_clean
            return Section(
                profile_name,
                "BG",
                h=h,
                w_btn=b,
                w_top=b,
                t_w=t1,
                t_fbtn=t4,
                t_ftop=t2,
                parent=fem,
            )
        elif sec_type.upper() == "CIRC":
            return Section(profile_name, "CIRC", r=props_clean[0], parent=fem)
        elif sec_type.upper() == "I":
            (
                l,
                h,
                b1,
                b2,
                t1,
                t2,
                t3,
            ) = props_clean
            return Section(
                profile_name,
                "IG",
                h=h,
                w_btn=b1,
                w_top=b2,
                t_w=t3,
                t_fbtn=t1,
                t_ftop=t2,
                parent=fem,
            )
        elif sec_type.upper() == "L":
            b, h, t1, t2 = props_clean
            return Section(profile_name, "HP", h=h, w_btn=b, t_w=t2, t_fbtn=t1, parent=fem)
        elif sec_type.upper() == "PIPE":
            r, t = props_clean
            return Section(profile_name, "TUB", r=r, wt=t, parent=fem)
        elif sec_type.upper() == "TRAPEZOID":
            # Currently converts Trapezoid to general beam
            b, h, a, d = props_clean
            # Assuming the Abaqus trapezoid element is symmetrical
            c = (b - a) / 2

            # The properties were quickly copied from a resource online. Most likely it contains error
            # https: // www.efunda.com / math / areas / trapezoidJz.cfm
            genprops = GeneralProperties(
                ax=h * (a + b) / 2,
                ix=h
                * (
                    b * h ** 2
                    + 3 * a * h ** 2
                    + a ** 3
                    + 3 * a * c ** 2
                    + 3 * c * a ** 2
                    + b ** 3
                    + c * b ** 2
                    + a * b ** 2
                    + b * c ** 2
                    + 2 * a * b * c
                    + b * a ** 2
                ),
                iy=(h ** 3) * (3 * a + b) / 12,
                iz=h
                * (
                    a ** 3
                    + 3 * a * c ** 2
                    + 3 * c * a ** 2
                    + b ** 3
                    + c * b ** 2
                    + a * b ** 2
                    + 2 * a * b * c
                    + b * a ** 2
                )
                / 12,
            )
            return Section(profile_name, "GENBEAM", genprops=genprops, parent=fem)
        else:
            raise ValueError(f'Currently unsupported section type "{sec_type}"')

    def grab_beam(match):
        d = match.groupdict()
        elset = fem.elsets[d["elset"]]
        name = d["sec_name"] if d["sec_name"] is not None else elset.name
        profile = d["profile_name"] if d["profile_name"] is not None else elset.name
        ass = fem.parent.get_assembly()
        material = ass.materials.get_by_name(d["material"])
        # material = parent.parent.materials.get_by_name(d['material'])
        temperature = d["temperature"]
        section_type = d["sec_type"]
        geo_props = d["line1"]
        sec = interpret_section(profile, section_type, geo_props)
        beam_y = [float(x.strip()) for x in d["line2"].split(",") if x.strip() != ""]
        metadata = dict(
            temperature=temperature,
            profile=profile.strip(),
            section_type=section_type,
            line1=geo_props,
        )
        res = fem.parent.sections.add(sec)
        if res is not None:
            sec = res
        return FemSection(
            name.strip(),
            sec_type="beam",
            elset=elset,
            section=sec,
            local_y=beam_y,
            material=material,
            metadata=metadata,
            parent=fem,
        )

    return map(grab_beam, re_beam.finditer(bulk_str))


def get_solid_sections_from_inp(bulk_str, fem):
    """

    ** Section: Section-80-MAT2TH1
    *Shell Section, elset=MAT2TH1, material=S3_BS__S355_16_T__40_M2
    0.02, 5

    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM

    """
    secnames = Counter(1, "solidsec")

    if bulk_str.lower().find("*solid section") == -1:
        return []
    re_solid = re.compile(
        r"(?:\*\s*Section:\s*(.*?)\n|)\*\Solid Section,\s*elset=(.*?)\s*,\s*material=(.*?)\s*$",
        _re_in,
    )
    solid_iter = re_solid.finditer(bulk_str)

    def grab_solid(m_in):
        name = m_in.group(1) if m_in.group(1) is not None else next(secnames)
        elset = m_in.group(2)
        material = m_in.group(3)
        return FemSection(
            name=name,
            sec_type="solid",
            elset=elset,
            material=material,
            parent=fem,
        )

    return map(grab_solid, solid_iter)


def get_shell_sections_from_inp(bulk_str, fem):
    """

    ** Section: Section-80-MAT2TH1
    *Shell Section, elset=MAT2TH1, material=S3_BS__S355_16_T__40_M2
    0.02, 5
    :param bulk_str:
    :param fem:
    :type fem: ada.fem.FEM
    :return: map object containing list of FemSection objects

    """

    shname = Counter(1, "sh")

    if bulk_str.lower().find("*shell section") == -1:
        return []
    re_offset = r"(?:, offset=(?P<offset>.*?)|)"
    re_controls = r"(?:, controls=(?P<controls>.*?)|)"

    re_shell = re.compile(
        rf"(?:\*\s*Section:\s*(?P<name>.*?)\n|)\*\Shell Section, elset"
        rf"=(?P<elset>.*?)\s*, material=(?P<material>.*?){re_offset}{re_controls}\s*\n(?P<t>.*?),"
        rf"(?P<int_points>.*?)$",
        _re_in,
    )

    def grab_shell(m):
        d = m.groupdict()
        name = d["name"] if d["name"] is not None else next(shname)
        elset = fem.sets.get_elset_from_name(d["elset"])
        material = d["material"]
        thickness = float(d["t"])
        offset = d["offset"]
        int_points = d["int_points"]
        metadata = dict(controls=d["controls"])
        return FemSection(
            name=name,
            sec_type="shell",
            thickness=thickness,
            elset=elset,
            material=material,
            int_points=int_points,
            offset=offset,
            parent=fem,
            metadata=metadata,
        )

    return map(grab_shell, re_shell.finditer(bulk_str))


def get_interactions_from_bulk_str(bulk_str, assembly):
    """

    :param bulk_str:
    :param assembly:
    :type assembly: ada.Assembly
    :return:
    """
    gen_name = Counter(1, "general")

    if bulk_str.find("** Interaction") == -1 and bulk_str.find("*Contact") == -1:
        return

    def resolve_surface_ref(surf_ref):
        surf_name = surf_ref.split(".")[-1] if "." in surf_ref else surf_ref
        surf = None
        if surf_name in assembly.fem.surfaces.keys():
            surf = assembly.fem.surfaces[surf_name]

        for p in assembly.get_all_parts_in_assembly():
            if surf_name in p.fem.surfaces.keys():
                surf = p.fem.surfaces[surf_name]
        if surf is None:
            raise ValueError("Unable to find surfaces in assembly parts")

        return surf

    for m in AbaCards.contact_pairs.regex.finditer(bulk_str):
        d = m.groupdict()
        intprop = assembly.fem.intprops[d["interaction"]]
        surf1 = resolve_surface_ref(d["surf1"])
        surf2 = resolve_surface_ref(d["surf2"])

        assembly.fem.add_interaction(Interaction(d["name"], "surface", surf1, surf2, intprop, metadata=d))

    for m in AbaCards.contact_general.regex.finditer(bulk_str):
        s = m.start()
        e = m.endpos
        interact_str = bulk_str[s:e]
        d = m.groupdict()
        intprop = assembly.fem.intprops[d["interaction"]]
        # surf1 = resolve_surface_ref(d["surf1"])
        # surf2 = resolve_surface_ref(d["surf2"])

        assembly.fem.add_interaction(
            Interaction(next(gen_name), "general", None, None, intprop, metadata=dict(aba_bulk=interact_str))
        )


def list_cleanup(membulkstr):
    return membulkstr.replace(",\n", ",").replace("\n", ",")


def is_set_in_part(part, set_name, set_type):
    """

    :param part:
    :param set_name:
    :param set_type:
    :type part: ada.Part
    :return: Set (node, element or surface)
    """

    set_map = {"nset": part.fem.nsets, "elset": part.fem.elsets, "surface": part.fem.surfaces}
    el_map = {"nset": part.fem.nodes, "elset": part.fem.elements}

    if set_name in set_map[set_type].keys():
        return set_map[set_type][set_name]
    else:
        _id = int(set_name)
        return el_map[set_type].from_id(_id)


def grab_set_from_assembly(set_str, fem, set_type):
    """

    :param set_str:
    :param fem:
    :param set_type:
    :type fem: ada.fem.FEM
    :rtype: Union[ada.fem.FemSet, ada.fem.Surface]
    """
    res = set_str.split(".")
    if len(res) == 1:
        set_map = {"nset": fem.nsets, "elset": fem.elsets, "surface": fem.surfaces}
        set_name = res[0]
        return set_map[set_type][set_name]
    else:
        set_name = res[1]
        p_name = res[0]
        for part in fem.parent.get_all_parts_in_assembly():
            if p_name == part.fem.instance_name:
                r = is_set_in_part(part, set_name, set_type)
                if r is not None:
                    return r
    raise ValueError(f'No {set_type} "{set_str}" was found')
